#!/usr/bin/python
from __future__ import print_function, absolute_import

#######################################################################
# Runs crest on an xyz file, can add in options to the code if needed #
#######################################################################

# Python Libraries
import os
from rdkit.Chem import rdMolTransforms
import glob
import subprocess
import rdkit
from rdkit.Chem import AllChem


def atompairs(mol, atom1, atom2, constraints):
    """returns addtional constraints
    Parameters
    ----------
    mol : RDKit mol object
            Mol object
    atom1 : str
            type of atom 1
    atom2 : str
            type of atom 2
    Returns
    -------
    [int,int] [int,int]
            atom pairs, constraint value
    """
    active = []
    for x in constraints:
        active.append(x[:2])

    for i, x in enumerate(active):
        active[i] = [int(j) for j in x]
    pairs = []
    bonds = [(x.GetBeginAtomIdx(), x.GetEndAtomIdx()) for x in mol.GetBonds()]
    for [a, b] in bonds:
        if [a + 1, b + 1] not in active and [b + 1, a + 1] not in active:
            dist = round(rdMolTransforms.GetBondLength(mol.GetConformer(), a, b), 3)
            at_a, at_b = mol.GetAtoms()[a].GetSymbol(), mol.GetAtoms()[b].GetSymbol()
            if atom1 == "X" and atom2 == "X":
                pairs.append([float(a + 1), float(b + 1), dist])
            elif atom1 == "X" and atom2 == "H":
                if at_a == "H" or at_b == "H":
                    pairs.append([float(a + 1), float(b + 1), dist])
            else:
                if (at_a == atom1 and at_b == atom2) or (
                    at_a == atom2 and at_b == atom1
                ):
                    pairs.append([float(a + 1), float(b + 1), dist])

    return pairs


def get_constraint(mol, constraints):
    # constrained optimization with xtb
    xx_pairs = atompairs(mol, "X", "X", constraints)
    # xh_pairs, xh_vals = atompairs(rdmol, "X", "H", active)
    # print(active, template)
    all_fix = []
    for x in constraints:
        all_fix.append(x)
    for x in xx_pairs:
        all_fix.append(x)
    return all_fix


def xyzall_2_xyz(xyzin, name):
    # converting multiple xyz to single
    command_run_1 = ["obabel", xyzin, "-oxyz", "-O" + name + "_conf_.xyz", "-m"]
    subprocess.run(command_run_1)


def run_xtb(
    xyzin, xyzoutxtb, constraints_dist, constraints_angle, constraints_dihedral, charge
):
    # check whether job has already been run
    if os.path.exists(xyzoutxtb):
        print("   {0} already exists: skipping xTB optimization".format(xyzoutxtb))
        # pass
    else:
        command = [
            os.path.abspath(os.path.dirname(__file__)) + "/run_xtb.sh",
            xyzin,
            "--xyzout",
            str(xyzoutxtb),
            "--charge",
            str(charge),
        ]

        if constraints_dist is not None:
            for i, bond in enumerate(constraints_dist):
                command.append("--dist")
                command.append(
                    str(int(bond[0])) + "," + str(int(bond[1])) + "," + str(bond[2])
                )
        if constraints_angle is not None:
            for i, angle in enumerate(constraints_angle):
                command.append("--angle")
                command.append(
                    str(int(angle[0]))
                    + ","
                    + str(int(angle[1]))
                    + ","
                    + str(int(angle[2]))
                    + ","
                    + str(angle[3])
                )
        if constraints_dihedral is not None:
            for i, dihedral in enumerate(constraints_dihedral):
                command.append("--dihedral")
                command.append(
                    str(int(dihedral[0]))
                    + ","
                    + str(int(dihedral[1]))
                    + ","
                    + str(int(dihedral[2]))
                    + ","
                    + str(int(dihedral[3]))
                    + ","
                    + str(dihedral[4])
                )
        subprocess.run(command)


def crest_opt(mol, name, dup_data, dup_data_idx, sdwriter, args, log):

    """run xtb using shell script and args
    Parameters
    ----------
    mol : RDKit Mol
            RDkit mol object
    name : str
            Name of molecule
    dup_data : pandas df
            csv
    dup_data_idx : int
            index of pandas df
    sdwriter : Rdkit writer
            sdwriter to write sdf file
    args : pyconfot args
            arguments for crest
    log : Logger
            Logging
    Returns
    -------
    performs crest/cregen conformer search
    """

    xyzin = name + ".xyz"

    if not args.ts_complex:
        AllChem.EmbedMolecule(mol)
        rdkit.Chem.rdmolfiles.MolToXYZFile(mol, xyzin)

    xyzoutxtb1 = name + "_xtb1.xyz"
    xyzoutxtb2 = name + "_xtb2.xyz"
    xyzoutall = name + "_conformers.xyz"
    xyzoutbest = name + "_best.xyz"
    charge = args.charge[0]
    mult = args.mult
    cbonds = args.cbonds
    constraints_dist = args.constraints_dist
    constraints_angle = args.constraints_angle
    constraints_dihedral = args.constraints_dihedral
    cregen = args.cregen
    cregen_ethr = args.cregen_ethr
    cregen_rthr = args.cregen_rthr
    cregen_bthr = args.cregen_bthr
    cregen_ewin = args.cregen_ewin

    all_fix = get_constraint(mol, constraints_dist)
    run_xtb(xyzin, xyzoutxtb1, all_fix, constraints_angle, constraints_dihedral, charge)
    run_xtb(
        xyzoutxtb1,
        xyzoutxtb2,
        constraints_dist,
        constraints_angle,
        constraints_dihedral,
        charge,
    )

    if os.path.exists(xyzoutall):
        print("   {0} already exists: skipping crest search".format(xyzoutall))
        # pass
    else:
        unique_atoms = []
        if constraints_dist is not None:
            for x in constraints_dist:
                for i in x[:2]:
                    if i not in unique_atoms:
                        unique_atoms.append(int(i))
        if constraints_angle is not None:
            for x in constraints_angle:
                for i in x[:3]:
                    if i not in unique_atoms:
                        unique_atoms.append(int(i))
        if constraints_dihedral is not None:
            for x in constraints_dihedral:
                for i in x[:4]:
                    if i not in unique_atoms:
                        unique_atoms.append(int(i))
        command = [
            os.path.abspath(os.path.dirname(__file__)) + "/run_crest.sh",
            xyzoutxtb2,
            "--xyzoutall",
            str(xyzoutall),
            "--xyzoutbest",
            str(xyzoutbest),
            "--charge",
            str(charge),
        ]
        if len(unique_atoms) != 0:
            command.append("--constraint")
            toadd = ",".join([str(elem) for elem in unique_atoms])
            command.append(toadd)
        if cbonds is not None:
            command.append("--cbonds")
            command.append(str(cbonds))

        subprocess.run(command)

    if cregen:
        xyzcregenensemble = xyzoutall.split(".xyz")[0] + "_cregen.xyz"
        xyzcregenbest = xyzoutbest.split(".xyz")[0] + "_cregen.xyz"

        # check whether job has already been run
        if os.path.exists(xyzcregenensemble):
            print(
                "   {0} already exists: skipping cregen search".format(
                    xyzcregenensemble
                )
            )
            # pass
        else:
            command = [
                os.path.abspath(os.path.dirname(__file__)) + "/run_cregen.sh",
                xyzoutall,
                xyzoutbest,
                "--xyzout",
                str(xyzcregenensemble),
                "--charge",
                str(charge),
                "--ethr",
                str(cregen_ethr),
                "--rthr",
                str(cregen_rthr),
                "--bthr",
                str(cregen_bthr),
                "--ewin",
                str(cregen_ewin),
            ]
            subprocess.run(command)

            # converting multiple xyz to single
            command_run_1 = [
                "obabel",
                xyzcregenensemble,
                "-oxyz",
                "-O" + xyzcregenbest,
                "-l",
                "1",
            ]
            subprocess.run(command_run_1)

    if cregen:
        xyzall_2_xyz(xyzcregenensemble, name)
    else:
        xyzall_2_xyz(xyzoutall, name)

    xyz_files = glob.glob(name + "_conf_*.xyz")
    for i, file in enumerate(xyz_files):
        name1 = file.split(".xyz")[0]
        command_xyz = ["obabel", "-ixyz", file, "-osdf", "-O" + name1 + ".sdf"]
        subprocess.call(command_xyz)
        os.remove(file)

    sdf_files = glob.glob(name + "*.sdf")
    for file in sdf_files:
        mol = rdkit.Chem.SDMolSupplier(file, removeHs=False)
        mol_rd = rdkit.Chem.RWMol(mol[0])
        energy = str(open(file, "r").readlines()[0])
        mol_rd.SetProp("Energy", energy)
        mol_rd.SetProp("Real charge", str(charge))
        sdwriter.write(mol_rd)
        os.remove(file)

    dup_data.at[dup_data_idx, "crest-conformers"] = len(xyz_files)

#     for f in glob.glob(os.getcwd() + "/" + name + "*.xyz"):
#         os.remove(f)

    return 1



######################################################.
#        This file stores all the functions          #
#          used in the LOG file analyzer             #
######################################################.
import os
import sys
import glob
import pandas as pd
import json
import subprocess
from pathlib import Path

from aqme.utils import periodic_table,get_info_input

from aqme.utils import (
	move_file,
    Logger,
    load_from_yaml,
    check_isomerization,
    read_file,
    cclib_atoms_coords,
)

from aqme.qprep import qprep
from aqme.argument_parser import set_options


class qcorr():
	"""
	Class containing all the functions from the QCORR module.

	Parameters
	----------
	qm_files : list 
		Filenames of QM output files to analyze
	w_dir_main : str
		Working directory
	dup_threshold : float
		Energy (in hartree) used as the energy difference in E, H and G to detect duplicates
	mem : str
		Memory for the QM calculations (i) Gaussian: total memory; (ii) ORCA: memory per processor
	nprocs : int
		Number of processors used in the QM calculations
	chk : bool
		Include the chk input line in new input files for Gaussian calculations
	qm_input : str
		Keywords line for new input files
	s2_threshold :  float
		Cut off for spin contamination during analysis in \%\ of the expected value 
		(i.e. multiplicity 3 has an the expected <S**2> of 2.0, if s2_threshold = 10,
		the <S**2> value is allowed to be 2.0 +- 0.2). Set s2_threshold = 0 to deactivate this option.
	isom : str
		Check for isomerization from the initial input file to the resulting output files. 
		It requires the extension of the initial input files (i.e. isom='com') and the folder of
		the input files must be added in the isom_inputs option
	isom_inputs : str
		Folder containing the initial input files to check for isomerization
	vdwfrac : float
		Fraction of the summed VDW radii that constitutes a bond between two atoms in the isomerization filter
	covfrac : float
		Fraction of the summed covalent radii that constitutes a bond between two atoms in the isomerization filter
	bs_gen : str
		Basis set used for gen(ECP) atoms	
	bs : str
		Basis set used for non gen(ECP) atoms in gen(ECP) calculations
	gen_atoms : list of str
		Atoms included in the gen(ECP) basis set
	qm_end : str
		Final line in the new input files
	amplitude_ifreq : float
		Amplitude used to displace the imaginary frequencies to fix
	freq_conv : str
		If a string is defined, it will remove calculations that converged during optimization but did not
		convergence in the subsequent frequency calculation. Options: opt sections as strings i.e. 
		(opt=(calcfc,maxstep=5)). If readfc is specified in the string, the chk option must be included as well.
		Turn this option off by using freq_conv=False.
	ifreq_cutoff : float
		Cut off for to consider whether a frequency is imaginary (absolute of the specified value is used)
	fullcheck : bool
		Perform an analysis to detect whether the calculations were done homogeneously
		(i.e. same level of theory, solvent, grid size, etc)
	author : str
		Author of the calculations
	program : str
		Program required to create the new input files
	varfile : str
		Option to parse the variables using a yaml file (specify the filename)
	kwargs : argument class
		Specify any arguments from the QCORR module
	"""
	
	def __init__(self, qm_files=[], w_dir_main=os.getcwd(), dup_threshold=0.0001,
				mem='4GB', nprocs=2, chk=False, qm_input='', s2_threshold=10.0, 
				isom=False, isom_inputs=os.getcwd(), vdwfrac=0.50, covfrac=1.10, 
				bs_gen='', bs='', gen_atoms=[], qm_end='', amplitude_ifreq=0.2, freq_conv='opt=(calcfc,maxstep=5)',
				ifreq_cutoff=0.0, fullcheck=True, author='', program='gaussian', varfile=None, **kwargs):
		
		self.initial_dir = Path(os.getcwd())
		self.w_dir_main = Path(w_dir_main)
		self.dup_threshold = dup_threshold
		self.mem = mem
		self.nprocs = nprocs
		self.chk = chk
		self.amplitude_ifreq = amplitude_ifreq
		self.freq_conv = freq_conv
		self.ifreq_cutoff = ifreq_cutoff
		self.qm_input = qm_input
		self.s2_threshold = s2_threshold
		self.fullcheck = fullcheck
		self.bs_gen = bs_gen
		self.bs = bs
		self.gen_atoms = gen_atoms
		self.qm_end = qm_end
		self.author = author
		self.program = program
		self.isom = isom
		self.isom_inputs = Path(isom_inputs)
		self.vdwfrac = vdwfrac
		self.covfrac = covfrac
		
		if "options" in kwargs:
			self.args = kwargs["options"]
		else:
			self.args = set_options(kwargs)

		self.args.varfile = varfile

		if varfile is not None:
			self.args, self.log = load_from_yaml(self.args, self.log)
			self.w_dir_main = Path(self.args.w_dir_main)
			self.qm_files = self.args.qm_files
			self.dup_threshold = self.args.dup_threshold
			self.mem = self.args.mem
			self.nprocs = self.args.nprocs
			self.chk = self.args.chk
			self.amplitude_ifreq = self.args.amplitude_ifreq
			self.freq_conv = self.args.freq_conv
			self.ifreq_cutoff = self.args.ifreq_cutoff
			self.qm_input = self.args.qm_input
			self.s2_threshold = self.args.s2_threshold
			self.fullcheck = self.args.fullcheck
			self.gen_atoms = self.args.gen_atoms
			self.bs_gen = self.args.bs_gen
			self.bs = self.args.bs
			self.qm_end = self.args.qm_end
			self.program = self.args.program
			self.author = self.args.author
			self.isom = self.args.isom
			self.isom_inputs = Path(self.args.isom_inputs)
			self.vdwfrac = self.args.vdwfrac
			self.covfrac = self.args.covfrac

		# go to working folder and detect QM output files
		os.chdir(self.w_dir_main)
		if isinstance(qm_files, list): 
			self.qm_files = qm_files
		else:
			self.qm_files = glob.glob(qm_files)

		# detects cycle of analysis (0 represents the starting point)
		self.round_num = check_run(w_dir_main)

		# start a log file to track the QCORR module
		error_setup = False
		try:
			self.log = Logger(self.w_dir_main / 'QCORR-run',f'{str(self.round_num)}')
		except FileNotFoundError:
			print('x  The PATH specified as input in the w_dir_main option might be invalid!')
			error_setup = True

		if len(self.qm_files) == 0 and not error_setup:
			print(f'x  There are no output files in {self.w_dir_main}.')
			self.log.write(f'x  There are no output files in {self.w_dir_main}.')
			self.log.finalize()
			move_file(self.w_dir_main.joinpath('dat_files/'), self.w_dir_main,f'QCORR-run_{str(self.round_num)}.dat')
			error_setup = True

		if error_setup:
			# this is added to avoid path problems in jupyter notebooks
			os.chdir(self.initial_dir)
			sys.exit()

		self.log.write(f"o  Analyzing output files in {self.w_dir_main}\n")
		self.qcorr_processing()

		self.log.finalize()
		move_file(self.w_dir_main.joinpath('dat_files/'), self.w_dir_main,f'QCORR-run_{str(self.round_num)}.dat')
		
		# this is added to avoid path problems in jupyter notebooks
		os.chdir(self.initial_dir)

	def qcorr_processing(self):
		"""
		General function of the QCORR module that:
		1. Analyzes the QM output files and moves output files with normal termination 
		and no extra imaginary frequencies to the same folder
		2. Generates input files to fix errors and extra imaginary frequencies
		3. Generates input files with new keywords lines from the normally terminated 
		files from point 1 (i.e. single-point energy corrections). Optionally, 
		the analysis from points 1 and 2  might be disabled with the nocheck=True or --nocheck option.
		"""

		file_terms = {'finished': 0, 'sp_calcs' : 0, 'extra_imag_freq': 0, 'ts_no_imag_freq': 0, 
			'freq_no_conv': 0, 'spin_contaminated': 0, 'duplicate_calc': 0, 'atom_error': 0,
			'scf_error': 0, 'before_E_error': 0, 'not_specified': 0,
			'geom_rules_qcorr': 0, 'isomerized': 0}   
		
		E_dup_list, H_dup_list, G_dup_list = [],[],[]

		for file in self.qm_files:

			file_name = file.split('.')[0]

			# create a json file with cclib and load a dictionary. This protocol is
			# favored over the traditional ccread since more data will be added to
			# the json file at the end
			command_run_1 = ['ccwrite', 'json', file]
			subprocess.run(command_run_1)
			cclib_incomp = False
			try:
				with open(file_name+'.json') as json_file:
					cclib_data = json.load(json_file)
			except FileNotFoundError:
				cclib_incomp = True
			if cclib_incomp:
				print(f'x  Potential cclib compatibility problem with file {file}')
				self.log.write(f'x  Potential cclib compatibility problem with file {file}')
				self.log.finalize()
				move_file(self.w_dir_main.joinpath('dat_files/'), self.w_dir_main,f'QCORR-run_{str(self.round_num)}.dat')

				# this is added to avoid path problems in jupyter notebooks
				os.chdir(self.initial_dir)
				sys.exit()

			# get number of atoms, multiplicity and number of imaginary freqs
			n_atoms = cclib_data['properties']['number of atoms']
			charge = cclib_data['properties']['charge']
			mult = cclib_data['properties']['multiplicity']

			# determine the calculation type (ground or transition state), keyword line,
			# grid size, and more parameters reading part of the output QM file
			outlines, outfile = read_file(self.w_dir_main,file)
			
			keywords_line,calc_type,mem,nprocs,qm_program,author,qm_solv,qm_emp = self.get_init_info(outlines)

			grid_size,s2_operator,s2_preannhi,level_of_theory,creation_date = self.grid_s2_info(outlines,qm_program)

			# determine error/unfinished vs normal terminations
			try:
				# freq for normal terminations
				if n_atoms == 1:
					freqs = []
					freq_displacements = []
				else:
					freqs = cclib_data['vibrations']['frequencies']
					freq_displacements = cclib_data['vibrations']['displacement']
				termination = 'normal'
				errortype = 'none'
				
				# spin contamination analysis using user-defined thresholds
				unpaired_e = mult-1
				# this first part accounts for singlet diradicals (threshold is 10% of the spin before annihilation)
				if unpaired_e == 0 and s2_operator != 0:
					if s2_operator > abs(self.s2_threshold/100)*s2_preannhi:
						errortype = 'spin_contaminated'
				else:
					spin = unpaired_e*0.5
					s2_expected_value = spin*(spin+1)
					spin_diff = abs(float(s2_operator)-s2_expected_value)
					if spin_diff > abs(self.s2_threshold/100)*s2_expected_value:
						errortype = 'spin_contaminated'

			except (AttributeError,KeyError):
				try:
					# if the optimization finished, only a freq job is required
					if cclib_data['optimization']['done']:
						termination = 'other'
						errortype = 'no_freq'
				except (AttributeError,KeyError):
					try:
						# single-point calcs (normal terminations with no opt)
						cclib_data['optimization']['geometric values']
						termination = 'other'
						errortype = 'not_specified'
					except (AttributeError,KeyError):
						try:
							# general errors
							cclib_data['atoms']['elements']['number']
							termination = 'normal'
							errortype = 'sp_calc'
							calc_type = 'SP calculation'
						except (AttributeError,KeyError):
							# the program lost the molecular information (i.e. crashed at the start)
							termination = 'other'
							errortype = 'before_E_error'
							
			# use very short reversed loop to find basis set incompatibilities
			if errortype in ['not_specified','sp_calc']:
				for i in reversed(range(len(outlines)-15,len(outlines))):
					if outlines[i-1].find('Atomic number out of range') > -1 or outlines[i-1].find('basis sets are only available') > -1:
						errortype = 'atomicbasiserror'
						break	
					if outlines[i].find('SCF Error') > -1:
						errortype = 'SCFerror'
						break
					if outlines[i].find('Normal termination') > -1:
						break

			# check for undesired imaginary freqs and data used by GoodVibes
			if termination == 'normal':
				atom_types,cartesians = cclib_atoms_coords(cclib_data)
				if errortype == 'none':
					# in eV, converted to hartree using the conversion factor from cclib
					E_dup = cclib_data['properties']['energy']['total']
					E_dup = E_dup/27.21138505
					# in hartree
					H_dup = cclib_data['properties']['enthalpy']
					G_dup = cclib_data['properties']['energy']['free energy']

					# detects if this calculation is a duplicate
					for i,_ in enumerate(E_dup_list):
						if abs(E_dup - E_dup_list[i]) < abs(self.dup_threshold):
							if abs(H_dup - H_dup_list[i]) < abs(self.dup_threshold):
								if abs(G_dup - G_dup_list[i]) < abs(self.dup_threshold):                        
									errortype = 'duplicate_calc'

				if errortype == 'none': 
					E_dup_list.append(E_dup)
					H_dup_list.append(H_dup)
					G_dup_list.append(G_dup)

					symmno,point_group,roconst,rotemp,errortype = self.symm_rot_data(outlines,qm_program,errortype)
					
					initial_ifreqs = 0
					for freq in freqs:
						if float(freq) < 0 and abs(float(freq)) > abs(self.ifreq_cutoff):
							initial_ifreqs += 1

					# exclude TS imag frequency
					if calc_type == 'transition_state':
						initial_ifreqs -= 1

					# gives new coordinates by displacing the normal mode(s) of the negative freq(s)
					if initial_ifreqs > 0:
						errortype = 'extra_imag_freq'
					
					elif initial_ifreqs < 0:
						errortype = 'ts_no_imag_freq'

					if errortype in ['extra_imag_freq','freq_no_conv']:
						if errortype == 'extra_imag_freq':
							cartesians = self.fix_imag_freqs(n_atoms, cartesians, freqs, freq_displacements, calc_type)
							# in case no previous OPT was done (only works if it's not a TS)
							opt_found = False
							for keyword in keywords_line.split():
								if keyword.lower().startswith('opt'):
									opt_found = True
							if not opt_found:
								keywords_line += ' opt'
								
						elif errortype == 'freq_no_conv':
							# adjust the keywords so only FREQ is calculated
							new_keywords_line = ''
							for keyword in keywords_line.split():
								if keyword.lower().startswith('opt'):
									keyword = self.freq_conv
								new_keywords_line += keyword
								new_keywords_line += ' '
							keywords_line = new_keywords_line

			# for calcs with finished OPT but no freqs
			elif termination != 'normal' and errortype not in ['ts_no_imag_freq','atomicbasiserror','before_E_error']:
				if errortype == 'no_freq':
					# adjust the keywords so only FREQ is calculated
					new_keywords_line = ''
					for keyword in keywords_line.split():
						if keyword.lower().startswith('opt'):
							keyword = ''
						else:
							new_keywords_line += keyword
							new_keywords_line += ' '
					keywords_line = new_keywords_line

					atom_types,cartesians = cclib_atoms_coords(cclib_data)

				else:
					# helps to fix SCF convergence errors
					if errortype == 'SCFerror':
						if keywords_line.find(' scf=xqc') > -1 or keywords_line.find(' scf=qc') > -1:
							new_keywords_line = ''
							for keyword in keywords_line.split():
								if keyword == 'scf=xqc':
									keyword = 'scf=qc'
								new_keywords_line += keyword
								new_keywords_line += ' '
							keywords_line = new_keywords_line
								
						else:
							keywords_line += ' scf=xqc'
					
					if errortype in ['not_specified','SCFerror']:
						RMS_forces = [row[1] for row in cclib_data['optimization']['geometric values']]
						min_RMS = RMS_forces.index(min(RMS_forces))
						
						atom_types,cartesians = [],[]
						per_tab = periodic_table()
						count_RMS = -1
						for i in range(60,len(outlines)):
							if outlines[i].find('Standard orientation:') > -1:
								count_RMS += 1
							if count_RMS == min_RMS:
								for j in range(i+5,i+5+n_atoms):
									massno = int(outlines[j].split()[1])
									if massno < len(per_tab):
										atom_symbol = per_tab[massno]
									else:
										atom_symbol = "XX"
									atom_types.append(atom_symbol)
									cartesians.append([float(outlines[j].split()[3]), float(outlines[j].split()[4]), float(outlines[j].split()[5])])
								break

			# check for isomerization
			if self.isom != False and errortype not in ['ts_no_imag_freq','atomicbasiserror','before_E_error']:
				isomerized = False
				init_csv = pd.DataFrame()
				os.chdir(self.isom_inputs)
				try:
					atoms_com, coords_com, atoms_and_coords = [],[],[]
					if len(self.isom.split('.')) == 1:
						atoms_and_coords,_ = get_info_input(f'{file_name}.{self.isom}')

					elif self.isom.split('.')[1] != 'csv':
						init_csv = pd.read_csv(self.args.isom)

					for line in atoms_and_coords:
						atoms_com.append(line.split()[0])
						coords_com.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])

					isomerized = check_isomerization(coords_com, cartesians, atoms_com, atom_types, self.vdwfrac, self.covfrac, init_csv, file)
				
				except FileNotFoundError:
					print(f'x  No com file were found for {file}, the check_geom test will be disabled for this calculation')
					self.log.write(f'x  No com file were found for {file}, the check_geom test will be disabled for this calculation')
				
				if isomerized:
					errortype = 'isomerization'
				os.chdir(self.w_dir_main)

			if errortype not in ['ts_no_imag_freq','atomicbasiserror','before_E_error','isomerization','duplicate_calc','spin_contaminated','none','sp_calc']:
				qprep(destination=Path(f'{os.getcwd()}/unsuccessful_QM_outputs/run_{self.round_num}/fixed_QM_inputs'), w_dir_main=self.w_dir_main,
					molecule=file_name, charge=charge, mult=mult,
					program=self.program, atom_types=atom_types,
					cartesians=cartesians, qm_input=keywords_line,
					mem=mem, nprocs=nprocs, chk=self.chk, qm_end=self.qm_end,
					bs_gen=self.bs_gen, bs=self.bs, gen_atoms=self.gen_atoms)

			print(f'{file}: Termination = {termination}, Error type = {errortype}')
			self.log.write(f'{file}: Termination = {termination}, Error type = {errortype}')

			# close the file
			outfile.close()

			# This part places the calculations in different folders depending on the type of termination
			file_terms,destination = self.organize_outputs(file,termination,errortype,file_terms)

			# write information about the QCORR analysis in a csv
			self.write_qcorr_csv(file_terms)

			# append new data from AQME into the json file from cclib and move json files
			# into the corresponding folders
			aqme_data = {'AQME data': {}}
			aqme_data['AQME data']['keywords line'] = keywords_line
			aqme_data['AQME data']['calculation type'] = calc_type
			aqme_data['AQME data']['memory'] = mem
			aqme_data['AQME data']['number of procs'] = nprocs
			aqme_data['AQME data']['QM program'] = qm_program
			aqme_data['AQME data']['author'] = author
			aqme_data['AQME data']['solvation'] = qm_solv
			aqme_data['AQME data']['dispersion'] = qm_emp
			aqme_data['AQME data']['grid'] = grid_size
			aqme_data['AQME data']['<S**2>'] = s2_operator
			aqme_data['AQME data']['level of theory'] = level_of_theory
			aqme_data['AQME data']['creation date'] =creation_date
			if termination == 'normal' and errortype == 'none':
				aqme_data['AQME data']['symmetry number'] = symmno
				aqme_data['AQME data']['point group'] = point_group
				aqme_data['AQME data']['rotational constants'] = roconst
				aqme_data['AQME data']['rotational temperatures'] = rotemp
			
			# update and move the json files to their corresponding destination
			with open(file_name+'.json', "r+") as json_file:
				data = json.load(json_file)
				data.update(aqme_data)
				json_file.seek(0)
				json.dump(data, json_file, indent=4, separators=(", ", ": "))

			destination_json = destination.joinpath('json_files/')
			move_file(destination_json, self.w_dir_main, file_name+'.json')

		if self.fullcheck:
			try: 
				destination_fullcheck = self.w_dir_main.joinpath('successful_QM_outputs/json_files/')
				json_files = glob.glob(f'{destination_fullcheck}/*.json')
			
				full_check(w_dir_main=destination_fullcheck,destination_fullcheck=destination_fullcheck,json_files=json_files)
			
			except FileNotFoundError:
				print('x  No normal terminations with no errors to run the full check analysis')
				self.log.write('x  No normal terminations with no errors to run the full check analysis')	
		
			
	# include geom filters (ongoing work)

	# 			if len(self.args.geom_rules) >= 1:
	# 				passing_rules = True
	# 				valid_mol_gen = True
	# 				self.log.write("  ----- geom_rules filter(s) will be applied to the output file -----\n")
	# 				try:
	# 					format_file = file.split('.')[1]
	# 					mol,ob_compat,rdkit_compat = output_to_mol(file,format_file,self.log)
	# 					print_error_geom_rules=False
	# 					if ob_compat and rdkit_compat:
	# 						passing_rules = geom_rules_output(mol,self.args,self.log,file,print_error_geom_rules)
	# 						if not passing_rules:
	# 							errortype = 'fail_geom_rules'
	# 					os.remove(file.split('.')[0]+'.mol')
	# 				except AttributeError:
	# 					valid_mol_gen = False
	# 					os.remove(file.split('.')[0]+'.mol')
	# 					self.log.write("The file could not be converted into a mol object, geom_rules filter(s) will be disabled\n")



	def get_init_info(self,outlines):
		"""
		Retrieves information from QM files that is not included in the cclib analysis. 


		Parameters
		----------
		outlines : list of str
			Lines of the QM output files

		Returns
		-------
		keywords_line : string
			Original keyword line (input) from the output QM file
		calc_type : str
			Type of the QM calculation (ground_state or transition_state)
		mem : str
			Memory used in the calculations
		nprocs : int
			Number of processors used in the calculations
		qm_program : str
			Program and version used in the calculation
		author : str
			Author of the calculations
		qm_solv : str
			Solvation model used in the calculations
		qm_emp : str
			Empirical model (if any) used in the calculations
		"""    

		end_keywords,keywords_line,author = False,'',''
		mem,nprocs,qm_program,skip_lines = '',0,'',0
		qm_solv,qm_emp = 'gas_phase','none'
		orca_solv_1, orca_solv_2, orca_solv_3 = 'gas phase','',''
		
		for i in range(0,len(outlines)):
			if outlines[i].find(' Gaussian ') > -1 and outlines[i].find('Revision') > -1:
				qm_program = outlines[i][1:-2]

			elif outlines[i].find('* O   R   C   A *') > -1:
				qm_program = 'ORCA'

			if qm_program.lower().find('gaussian') > -1:
				# retrieve multiple parameters of the calculation
				if outlines[i].startswith(' #'):
					keywords_line += outlines[i].rstrip("\n")
					for j in range(i+1,i+100):
						if outlines[j].find('----------') > -1:
							end_keywords = True
							break
						if not end_keywords:
							keywords_line += outlines[j].rstrip("\n")[1:]
					keywords_line = keywords_line[3:]

				elif outlines[i].find('%mem') > -1:
					mem = outlines[i].strip().split('=')[-1]

				elif outlines[i].find('%nprocs') > -1:
					nprocs = outlines[i].strip().split('=')[-1]
				
				if keywords_line != '':
					break
			
			elif qm_program.lower().find('orca') > -1:
				# get program version
				if outlines[i].find('Program Version') > -1:
					qm_program = "ORCA, version " + outlines[i].split()[2]
				
				# retrieve the input lines, mem and nprocs
				if outlines[i].startswith('NAME ='):
					for j in range(i+1,i+100):
						if outlines[j].find('xyz') > -1 and outlines[j].find('*') > -1:
							end_keywords = True
							break

						if not end_keywords:
							if outlines[j].lower().find('%maxcore') > -1:
								mem = outlines[j].split()[-1]+'MB'

							# multiple ways to describe %pal
							elif outlines[j].lower().find('%pal') > -1 and outlines[j].lower().find('nprocs') > -1 and outlines[j].lower().find('end') > -1:
								nprocs = int(outlines[j].split()[-2])
							elif outlines[j].lower().find('%pal') > -1 and outlines[j].lower().find('nprocs') > -1 and outlines[j+1].lower().find('end') > -1:
								nprocs = int(outlines[j].split()[-1])
								skip_lines = 1
							elif outlines[j].lower().find('%pal') > -1 and outlines[j+1].lower().find('nprocs') > -1 and outlines[j+1].lower().find('end') > -1:
								nprocs = int(outlines[j+1].split()[-2])
								skip_lines = 1
							elif outlines[j].lower().find('%pal') > -1 and outlines[j+1].lower().find('nprocs') > -1 and outlines[j+2].lower().find('end') > -1:
								nprocs = int(outlines[j+1].split()[-1])
								skip_lines = 2
							elif skip_lines != 0:
								skip_lines -= 1

							else:
								keywords_line += outlines[j][6:]

				# ORCA parsing for solvation model
				elif outlines[i].find('CPCM SOLVATION MODEL') > -1:
					orca_solv_1 = 'CPCM, '
				elif outlines[i].find('SMD-CDS solvent descriptors:') > -1:
					orca_solv_2 = 'SMD, '
				elif outlines[i].find('Solvent:       ') > -1:
					orca_solv_3 = outlines[i].strip().split()[-1].lower()

				if outlines[i].find('SCF ITERATIONS') > -1:
					break
		
		if self.author != '':
			author = self.author

		# user-defined keywords line, mem and nprics overwrites previously used parameters
		if self.qm_input != '':
			keywords_line = self.qm_input

		if self.mem != '':
			mem = self.mem
		elif mem == '':
			mem = '8GB'

		if self.nprocs != 0:
			nprocs = self.nprocs
		elif nprocs == 0:
			nprocs = 4

		# find if the calculation was for a ground or transition state and get individual keywords
		calc_type = 'ground_state'
		calcfc_found, ts_found = False, False

		if qm_program.lower().find('gaussian') > -1:
			for keyword in keywords_line.split():
				if keyword.lower().find('opt') > -1:
					if keyword.lower().find('calcfc') > -1:
						calcfc_found = True
					if keyword.lower().find('ts') > -1:
						ts_found = True
				elif keyword.lower().startswith('scrf'):
					qm_solv = keyword
				elif keyword.lower().startswith('emp'):
					qm_emp = keyword
			if calcfc_found and ts_found:
				calc_type = 'transition_state'

		
		elif qm_program.lower().find('orca') > -1:
			for keyword in keywords_line.split():
				if keyword.lower() in ['optts','scants']:
					ts_found = True
					if keyword.lower() == 'scants':
						keyword = 'OptTS'
			if ts_found:
				calc_type = 'transition_state'
			if orca_solv_1 == 'CPCM, ' and orca_solv_2 == 'SMD, ':
				qm_solv = orca_solv_2 + orca_solv_3
			else:		
				qm_solv = orca_solv_1 + orca_solv_2 + orca_solv_3
	
		return keywords_line,calc_type,mem,nprocs,qm_program,author,qm_solv,qm_emp


	def symm_rot_data(self,outlines,qm_program,errortype):
		"""
		Retrieves information from QM files regarding symmetry and rotational parameters used by GoodVibes.


		Parameters
		----------
		outlines : list of str
			Lines of the QM output files
		qm_program : str
			Program and version used in the calculation
		errortype : string
			Type of error type of the QM output file (i.e. None, not_specified, extra_imag_freq, etc)

		Returns
		-------
		symmno : int
			Rotational symmetry number
		point_group : int
			Symmetry point group
		roconst : list of float
			Rotational constants in GHz
		rotemp : list of float
			Rotational temperatures
		"""  
		
		symmno, point_group, roconst, rotemp = '','',[],[]
		for i in reversed(range(0,len(outlines)-15)):
			if qm_program.lower().find('gaussian') > -1:
				if outlines[i].find('Rotational symmetry number') > -1:
					symmno = int((outlines[i].strip().split()[3]).split(".")[0])

				elif outlines[i].find('Full point group') > -1:
					point_group = outlines[i].strip().split()[3]

				elif outlines[i].find('Rotational constants (GHZ):') > -1:
					try:
						roconst = [float(outlines[i].strip().replace(':', ' ').split()[3]),
										float(outlines[i].strip().replace(':', ' ').split()[4]),
										float(outlines[i].strip().replace(':', ' ').split()[5])]
					except ValueError:
						if outlines[i].find('********') > -1:
							roconst = [float(outlines[i].strip().replace(':', ' ').split()[4]),
											float(outlines[i].strip().replace(':', ' ').split()[5])]
				
				elif outlines[i].find('Rotational temperature ') > -1:
					rotemp = [float(outlines[i].strip().split()[3])]
				elif outlines[i].find('Rotational temperatures') > -1:
					try:
						rotemp = [float(outlines[i].strip().split()[3]), float(outlines[i].strip().split()[4]),
								  float(outlines[i].strip().split()[5])]
					except ValueError:
						if outlines[i].find('********') > -1:
							rotemp = [float(outlines[i].strip().split()[4]), float(outlines[i].strip().split()[5])]
							
				if point_group != '' and symmno != '' and roconst != [] and rotemp != []:
					break

				elif self.freq_conv and outlines[i].find('Converged?') > -1:
					for j in range(i+1,i+5):
						if outlines[j].strip().split()[-1] == 'NO':
							errortype = 'freq_no_conv'
					if errortype == 'freq_no_conv':
						break
				
				elif outlines[i].find('Normal termination') > -1:
					break

		return symmno,point_group,roconst,rotemp,errortype


	def grid_s2_info(self,outlines,qm_program):
		"""
		Checks multiple parameters from QM files that are not included in the cclib analysis. 


		Parameters
		----------
		outlines : list of str
			Lines of the QM output files
		qm_program : str
			Program and version used in the calculation

		Returns
		-------
		grid_size : str
			Grid size used in the calculations
		s2_operator : float
			<S**2> value from open-shell QM calculations
		level_of_theory : str
			Functional and basis set used in the QM calculations
		creation_date : str
			Date of creation of the QM calculations (when they finish)
		"""  

		grid_size,s2_operator,s2_preannhi,s2_found,level_of_theory,creation_date = '',0,0,False,'',''
		grid_lookup = {1: 'sg1', 2: 'coarse', 4: 'fine', 5: 'ultrafine', 7: 'superfine'}
		found_theory,level,bs = False,'',''
		line_options = ['\\freq\\','|freq|','\\sp\\','|sp|']

		for i in reversed(range(0,len(outlines))):
			# get grid size
			if qm_program.lower().find('gaussian') > -1:
				if outlines[i].find('ExpMin=') > -1:
					IRadAn = int(outlines[i].strip().split()[-3])
					grid_size = grid_lookup[IRadAn]
				elif outlines[i].find('S**2 before annihilation') > -1 and not s2_found:
					s2_operator = float(outlines[i].strip().split()[-1])
					s2_preannhi = float(outlines[i].strip().split()[-3][:-1])
					s2_found = True

				elif outlines[i].find('Normal termination') > -1:
					creation_date = ' '.join(item for item in outlines[i].strip().split()[-4:])
					creation_date = creation_date[:-1]

				# get functional and basis set
				if not found_theory:
					if outlines[i].find('External calculation') > -1:
						level, bs = 'ext', 'ext'
						level_of_theory = '/'.join([level, bs])
						found_theory = True
					if i < len(outlines)-1:
						for option in line_options:
							if option in (outlines[i].lower()+outlines[i+1].lower()):
								level, bs = ((outlines[i]+outlines[i+1]).strip().replace('|','\\').split("\\")[4:6])
								found_theory = True
					# Remove the restricted R or unrestricted U label
					if level != '':
						if level[0] in ('R', 'U'):
							level = level[1:]
						level_of_theory = '/'.join([level, bs])

			if level_of_theory != '':
				found_theory = True
			
			elif qm_program.lower().find('orca') > -1:
				if outlines[i].lower().find('grid') > -1 and outlines[i].startswith('|  '):
					for keyword in outlines[i].lower().split():
						if keyword.find('grid') > -1:
							if grid_size != '':
								grid_size += ' '
							grid_size += keyword
				elif outlines[i].find('xyz') > -1 and outlines[i].find('*') > -1:
					if grid_size == '':
						grid_size = 'default multigrid'

			if grid_size != '':
				break

		# designed to detect G4 calcs
		if level_of_theory == 'HF/GFHFB2':
			level_of_theory = 'G4'

		return grid_size,s2_operator,s2_preannhi,level_of_theory,creation_date


	def fix_imag_freqs(self, n_atoms, cartesians, freqs, freq_displacements, calc_type):
		"""
		Fixes undersired (extra) imaginary frequencies from QM calculations.
		This function multiplies the imaginary normal mode vectors by the selected amplitude 
		(0.2 is the default amplitude in the pyQRC script from GitHub, user: bobbypaton).
		By default, all the extra imaginary modes are used (i.e. in calculations with three
		extra imaginary frequencies, all the three modes will be used to displace the atoms).
		This can be tuned with the --ifreq_cutoff option (i.e. only use freqs lower than -50 cm-1).

		Parameters
		----------
		n_atoms : int
			Number of atoms in the calculation
		cartesians : list of lists
			List of lists containing the molecular coordinates as floats
		freqs : list of float
			List containing the frequencies as floats
		freq_displacements : list of matrixes
			Contains the normal modes for each frequencies (including all atoms)
		calc_type : str
			Type of the QM calculation (ground_state or transition_state)

		Returns
		-------
		cartesians : list of lists
			New set of cartesian coordinates generated after displacing the original
			coordinates along the normal modes of the corresponding imaginary frequencies
		"""

		shift = []

		# could get rid of atomic units here, if zpe_rat definition is changed
		for mode,_ in enumerate(freqs):
			# moves along all imaginary freqs (ignoring the TS imag freq, assumed to be the most negative freq)
			if mode == 0 and calc_type == 'transition_state':
				shift.append(0.0)
			else:
				if freqs[mode] < 0.0:
					shift.append(self.amplitude_ifreq)
				else:
					shift.append(0.0)

			# The starting geometry is displaced along each normal mode according to the random shift
			for atom in range(0,n_atoms):
				for coord in range(0,3):
					cartesians[atom][coord] = cartesians[atom][coord] + freq_displacements[mode][atom][coord] * shift[mode]

		return cartesians


	def organize_outputs(self,file,termination,errortype,file_terms):
		"""
		1. Moves the QM output files to their corresponding folders after the analysis. 
		2. Keeps track of the number of calculations with the different types 
		of terminations and error types

		Parameters
		----------
		file : str
			Output file
		termination : string
			Type of termination of the QM output file (i.e. normal, error, unfinished)
		errortype : string
			Type of error type of the QM output file (i.e. None, not_specified, extra_imag_freq, etc)
		file_terms : dictionary
			Keeps track of the number of calculations for each termination and error type

		Returns
		-------
		destination : str
			Path to store the file
		"""
		
		if errortype == 'none' and termination == "normal":
			destination = self.w_dir_main.joinpath('successful_QM_outputs/')
			file_terms['finished'] += 1

		elif errortype == 'sp_calc' and termination == "normal":
			destination = self.w_dir_main.joinpath('successful_QM_outputs/SP_calcs/')
			file_terms['sp_calcs'] += 1

		elif errortype == 'extra_imag_freq':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/extra_imag_freq/')
			file_terms['extra_imag_freq'] += 1

		elif errortype == 'ts_no_imag_freq':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/ts_no_imag_freq/')
			file_terms['ts_no_imag_freq'] += 1

		elif errortype == 'spin_contaminated':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/spin_contaminated/')
			file_terms['spin_contaminated'] += 1

		elif errortype == 'duplicate_calc':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/duplicates/')
			file_terms['duplicate_calc'] += 1

		elif errortype == "atomicbasiserror":
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/error/basis_set_error/')
			file_terms['atom_error'] += 1

		elif errortype == "SCFerror":
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/error/scf_error/')
			file_terms['scf_error'] += 1

		elif errortype == "before_E_calculation":
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/error/before_E_calculation/')
			file_terms['before_E_error'] += 1

		elif errortype == 'fail_geom_rules':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/geom_rules_filter/')
			file_terms['geom_rules_qcorr'] += 1

		elif errortype == 'isomerization':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/isomerization/')
			file_terms['isomerized'] += 1

		elif errortype == 'freq_no_conv':
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/freq_no_conv/')
			file_terms['freq_no_conv'] += 1

		else:
			destination = self.w_dir_main.joinpath(f'unsuccessful_QM_outputs/run_{self.round_num}/error/not_specified_error/')
			file_terms['not_specified'] += 1
		
		move_file(destination, self.w_dir_main, file)

		return file_terms,destination

					
	def write_qcorr_csv(self,file_terms):
		"""
		Write information about the QCORR analysis in a csv
		"""

		ana_data = pd.DataFrame()
		ana_data.at[0,'Total files'] = len(self.qm_files)
		ana_data.at[0,'Normal termination'] = file_terms['finished']
		ana_data.at[0,'Extra imag. freq.'] = file_terms['extra_imag_freq']
		ana_data.at[0,'TS with no imag. freq.'] = file_terms['ts_no_imag_freq']
		ana_data.at[0,'Freq not converged'] = file_terms['freq_no_conv']
		ana_data.at[0,'SCF error'] = file_terms['scf_error']
		ana_data.at[0,'Error before SCF'] = file_terms['before_E_error']
		ana_data.at[0,'Basis set error'] =  file_terms['atom_error']
		ana_data.at[0,'Other errors'] = file_terms['not_specified']
		if self.args.s2_threshold > 0.0:
			ana_data.at[0,'Spin contamination'] = file_terms['spin_contaminated']
		ana_data.at[0,'Duplicates'] = file_terms['duplicate_calc']
		if len(self.args.geom_rules) >= 1:
			ana_data.at[0,'geom_rules filter'] = file_terms['geom_rules_qcorr']
		if self.isom != False:
			ana_data.at[0,'Isomerization'] = file_terms['isomerized']
		path_as_str = self.w_dir_main.as_posix()
		ana_data.to_csv(path_as_str+f'/QCORR-run_{self.round_num}-stats.csv',index=False)

		move_file(self.w_dir_main.joinpath('dat_files/'), self.w_dir_main,f'QCORR-run_{self.round_num}-stats.csv')


def full_check(w_dir_main=os.getcwd(),destination_fullcheck='',json_files=glob.glob('*.json')):
	"""
	Checks that multiple calculations were done following the same protocols, including
	program and version, grid size, level of theory, dispersion and solvation model.

	Parameters
	----------
	w_dir_main : str
		Working directory
	destination_fullcheck : str
		Destination to create the file with the full check
	json_files : list of str
		json files to compare (glob.glob('*.json') and '*.json are both valid inputs to
		include all the json files from a folder)	
	"""

	initial_dir = os.getcwd()
	w_dir_main = Path(w_dir_main)
	os.chdir(w_dir_main)

	if json_files == '*.json' or json_files == '*json':
		json_files=glob.glob('*.json')

	df_fullcheck = pd.DataFrame(columns=['file', 'program', 'grid', 'lot', 'dispersion', 'solvation'])

	for file in json_files:
		file_name = file.split('.')[0]
		with open(file) as json_file:
			cclib_data = json.load(json_file)
		program = cclib_data['AQME data']['QM program']
		grid = cclib_data['AQME data']['grid']
		lot = cclib_data['AQME data']['level of theory']
		dispersion = cclib_data['AQME data']['dispersion']
		solvation = cclib_data['AQME data']['solvation']
		df_fullcheck.loc[len(df_fullcheck.index)] = [file_name, program, grid, lot, dispersion, solvation] 
	
	fullcheck_file = '--QCORR_Fullcheck_Analysis--.dat'
	fullcheck_txt = '-- Full check analysis --'

	for prop in df_fullcheck.columns:
		if prop != 'file':
			unique_props = df_fullcheck[prop].unique()
			if len(unique_props) > 1:
				fullcheck_txt += f'\nx  Different {prop} used in the calculations:'
				for unique_prop in unique_props:
					file_names = df_fullcheck["file"].loc[df_fullcheck[prop] == unique_prop]
					fullcheck_txt += f'\n     * {unique_prop} in:'
					for file_name in file_names:
						adapted_name = file_name.replace('/','\\').split("\\")[-1]
						fullcheck_txt += f'\n       - {adapted_name}'
			else:
				fullcheck_txt += f'\no  Same {prop} ({unique_props[0]}) used in all the calculations'
			
	fullcheck_analysis = open(fullcheck_file, 'w')
	print(fullcheck_txt)
	fullcheck_analysis.write(fullcheck_txt)	
	fullcheck_analysis.close()

	if destination_fullcheck == '':
		destination_fullcheck = w_dir_main.joinpath('successful_QM_outputs/json_files/')
	else:
		destination_fullcheck = Path(destination_fullcheck)
	move_file(destination_fullcheck, w_dir_main, fullcheck_file)
	
	os.chdir(initial_dir)


def json2input(json_files=[], w_dir_main=os.getcwd(), destination=None, suffix='', 
				charge=None, mult=None,	mem='8GB', nprocs=4, chk=False, qm_input='', bs_gen='', 
				bs='', gen_atoms=[], qm_end='', program='gaussian'):
	'''
	Reads a json file and use QPREP to generate input files.

	Parameters
	----------
	json_files : list 
		Filenames of json files to analyze
	w_dir_main : str
		Folder with the json files to process
	destination : str
		Destination to create the new input files
	suffix : str
		Suffix for the new input files
	charge : int
		Charge of the calculations used in the following input files
	mult : int
		Multiplicity of the calculations used in the following input files
	mem : str
		Memory used in the calculations
	nprocs : int
		Number of processors used in the calculations
	chk : bool
		Include the chk input line in new input files
	qm_input : str
		Keywords line for new input files
	bs_gen : str
		Basis set used for gen(ECP) atoms	
	bs : str
		Basis set used for non gen(ECP) atoms in gen(ECP) calculations
	gen_atoms : list of str
		Atoms included in the gen(ECP) basis set
	qm_end : str
		Final line in the new input files
	program : str
		Program required to create the new input files
	'''
	
	w_dir_initial=os.getcwd()
	os.chdir(w_dir_main)

	charge_initial = charge
	mult_initial = mult

	if destination is None:
		destination = Path(w_dir_main).joinpath("QCALC")
	else:
		destination = Path(destination)

	if not isinstance(json_files, list): 
		json_files = glob.glob(json_files)

	if json_files == '*.json':
		json_files = glob.glob('*.json')
	
	for file in json_files:
		file_name = file.split('.')[0]
		with open(file) as json_file:
			cclib_data = json.load(json_file)
		try:
			atom_types,cartesians = cclib_atoms_coords(cclib_data)
		except (AttributeError,KeyError):
			print('x  The json files do not contain coordinates and/or atom type information')
		
		# if no charge and multiplicity are specified, they are read from the json file
		if charge_initial is None:
			charge = cclib_data['properties']['charge']
		if mult_initial is None:
			mult = cclib_data['properties']['multiplicity']
		if charge is None:
			print('x  No charge was specified in the json file or function input (i.e. json2input(charge=0) )')
		elif mult is None:
			print('x  No multiplicity was specified in the json file or function input (i.e. json2input(mult=1) )')

		qprep(destination=destination, w_dir_main=w_dir_main,
			molecule=file_name, charge=charge, mult=mult,
			program=program, atom_types=atom_types,
			cartesians=cartesians, qm_input=qm_input,
			mem=mem, nprocs=nprocs, chk=chk, qm_end=qm_end,
			bs_gen=bs_gen, bs=bs, gen_atoms=gen_atoms, suffix=suffix)
	
	print(f'o  Final input files were generated in {destination}')

	os.chdir(w_dir_initial)
	

def check_run(w_dir):
	"""
	# Determines the folder where input files are gonna be generated in QCORR.
	"""

	input_folder = w_dir+'/unsuccessful_QM_outputs/'
	folder_count = 0

	if os.path.exists(input_folder):
		dir_list = os.listdir(input_folder)
		for folder in dir_list:
			if folder.find('run_') > -1:
				folder_count += 1

	return folder_count


#########################################################.
#        This file stores all the functions             #
#          used in writing SDF and COM files,           #
#              as well as the logger and                #
#                 yaml file importer                    #
#########################################################.

import os
import sys
import pandas as pd
from rdkit import Chem
from aqme.utils import load_from_yaml, Logger, move_file
from aqme.argument_parser import set_options
from openbabel import pybel
from pathlib import Path


# template classes
class qprep:

	"""
	Class containing all the functions from the QPREP module related to Gaussian input files.

	Parameters
	----------
	mol : mol object with 3D coordinates
		Optionally, starts from a mol object to prepare the input QM file
	w_dir_main : str
		Working directory
	destination : str
		Directory to create the input file
	molecule : str
		Name of the input file (without the extension i.e. NAME for NAME.com)
	charge : int
		Charge of the calculations used in the following input files
	mult : int
		Multiplicity of the calculations used in the following input files
	chk : bool
		True to include the initial %chk line in Gaussian
	mem : str
		Memory used in the input file. Formats: GB, MB or MW (i.e. 4GB, 800MB or 16MW).
	nprocs : int
		Number of processors used in the input file
	suffix : str
		Suffix for the new input files
	atom_types : list of strings
		(If no mol is used) List containing the atoms of the system
	cartesians : list of lists
		(If no mol is used) Cartesian coordinates used for further processing
	bs_gen : str
		Basis set for the Gen(ECP) section
	bs : str
		Basis set for regular atoms when Gen(ECP) is used
	gen_atoms : str
		Atoms considered for Gen(ECP). Format: ATOM1,ATOM2,ATOM3... (i.e. C,H,O)
	program : str
		Target program to generate input files. Options: 'gaussian' and 'orca'
	qm_input : string
		Keyword line of the input file
	qm_end : str
		Final line(s) of the input file
	varfile : str
		Option to parse the variables using a yaml file (specify the filename)
	kwargs : argument class
		Specify any arguments from the QCORR module
	"""

	def __init__(
		self,
		mol=None,
		w_dir_main=os.getcwd(),
		destination=None,
		molecule="",
		charge=0,
		mult=1,
		chk=False,
		mem="8GB",
		nprocs=4,
		suffix='',
		atom_types=[],
		cartesians=[],
		bs_gen="",
		bs="",
		gen_atoms=None,
		program="gaussian",
		qm_input="",
		qm_end="",
		varfile=None,
		**kwargs,
	):

		self.mol = mol
		self.w_dir_main = Path(w_dir_main)
		self.molecule = molecule
		self.chk = chk
		self.mem = mem
		self.nprocs = nprocs
		self.suffix = suffix
		self.charge = charge
		self.mult = mult
		self.qm_input = qm_input
		self.bs_gen = bs_gen
		self.bs = bs
		self.gen_atoms = gen_atoms
		self.qm_end = qm_end
		self.program = program

		if destination is None:
			self.destination = Path(os.getcwd()+"/QCALC/")
      
		else:
			self.destination = Path(destination)

		self.log = Logger(self.w_dir_main / 'QPREP', 'data')

		if "options" in kwargs:
			self.args = kwargs["options"]
		else:
			self.args = set_options(kwargs)

		self.args.varfile = varfile

		if varfile is not None:
			self.args, self.log = load_from_yaml(self.args, self.log)
			self.w_dir_main = Path(self.args.w_dir_main)
			self.destination = self.args.destination
			if destination is None:
				self.destination = self.w_dir_main.joinpath("QCALC")
			else:
				self.destination = Path(self.args.destination)
			self.charge = self.args.charge
			self.mult = self.args.mult
			if self.charge == None:
				self.charge = 0
			if self.mult == None:
				self.mult = 1
			self.chk = self.args.chk
			self.mem = self.args.mem
			self.nprocs = self.args.nprocs
			self.qm_input = self.args.qm_input
			self.gen_atoms = self.args.gen_atoms
			self.bs_gen = self.args.bs_gen
			self.bs = self.args.bs
			self.qm_end = self.args.qm_end
			self.program = self.args.program
			self.suffix = self.args.suffix

		self.atom_types = atom_types
		self.cartesians = cartesians

		if self.mol != None:
			self.atom_types = [ atom.GetSymbol() for _, atom in enumerate(self.mol.GetAtoms())]
			self.cartesians = self.mol.GetConformers()[0].GetPositions()
		self.n_atoms = len(self.atom_types)

		if self.qm_input == "":
			sys.exit("x  No keywords line was specified! (qm_input=KEYWORDS_LINE).")

		if molecule == "":
			sys.exit("x  No name was specified! (molecule=NAME).")
		
		comfile = self.write()
		move_file(self.destination, self.w_dir_main, comfile)
		self.log.finalize()
		os.remove('QPREP_data.dat')


	def get_header(self):
		'''
		Gets the part of the input file above the molecular coordinates.
		'''

		txt = ''

		if self.program.lower() == 'gaussian':
			if self.chk:
				txt += f'%chk={self.molecule}.chk\n'
			txt += f'%nprocshared={self.nprocs}\n'
			txt += f'%mem={self.mem}\n'
			txt += f'# {self.qm_input}'
			txt += f'\n\n'
			txt += f'{self.molecule}\n\n'
			txt += f'{self.charge} {self.mult}\n'

		elif self.program.lower() == 'orca':
			txt += f'# {self.molecule}\n'
			if self.mem.find('GB'):
				mem_orca = int(self.mem.split('GB')[0])*1000
			elif self.mem.find('MB'):
				mem_orca = self.mem.split('MB')[0]
			elif self.args.mem.find('MW'):
				mem_orca = self.mem.split('MW')[0]
			txt += f'%maxcore {mem_orca}\n'
			txt += f'%pal nprocs {self.nprocs} end\n'
			txt += f'! {self.qm_input}\n'
			txt += f'* xyz {self.charge} {self.mult}\n'

		return txt

	def get_tail(self):
		"""
		Gets the part of the input file below the molecular coordinates.
		"""

		txt = ""

		if self.program.lower() == "gaussian":
			if self.gen_atoms is not None and len(self.gen_atoms) > 0:
				# writes part for Gen/GenECP
				ecp_used, ecp_not_used, gen_type = [], [], "gen"
				if self.qm_input.lower().find("genecp") > -1:
					gen_type = "genecp"

				for _, element_ecp in enumerate(self.atom_types):
					if element_ecp in self.gen_atoms and element_ecp not in ecp_used:
						ecp_used.append(element_ecp)
					elif (
						element_ecp not in self.gen_atoms
						and element_ecp not in ecp_not_used
					):
						ecp_not_used.append(element_ecp)

				if len(ecp_not_used) > 0:
					elements_not_used = " ".join([f"{sym}" for sym in ecp_not_used])
					txt += f"{elements_not_used} 0\n{self.bs}\n****\n"
				if len(ecp_used) > 0:
					elements_used = " ".join([f"{sym}" for sym in ecp_used])
					txt += f"{elements_used} 0\n{self.bs_gen}\n****\n"

				if gen_type == "genecp" and len(ecp_used) > 0:
					txt += "\n"
					txt += f"{elements_used} 0\n{self.bs_gen}\n****\n"

				txt += "\n"

			# writes final section if selected
			if self.qm_end != "":
				txt += f"{self.qm_end}\n\n"

		return txt

	def write(self):

		if self.program.lower() == 'gaussian':
			extension = 'com'
		elif self.program.lower() == 'orca':
			extension = 'inp'
		if self.suffix != '':
			comfile = f'{self.molecule}_{self.suffix}.{extension}'
		else:
			comfile = f'{self.molecule}.{extension}'

		if os.path.exists(comfile):
			os.remove(comfile)

		header = self.get_header()
		tail = self.get_tail()

		fileout = open(comfile, "w")
		fileout.write(header)

		for atom in range(0, self.n_atoms):
			fileout.write(
				"{0:>2} {1:12.8f} {2:12.8f} {3:12.8f}".format(
					self.atom_types[atom],
					self.cartesians[atom][0],
					self.cartesians[atom][1],
					self.cartesians[atom][2],
				)
			)
			if atom != self.n_atoms - 1:
				fileout.write("\n")

		if self.program.lower() == "gaussian":
			fileout.write("\n\n")
		elif self.program.lower() == "orca":
			fileout.write("\n*")

		fileout.write(tail)
		fileout.close()

		return comfile


# Aux Functions for QM input generation
def get_molecule_list(
	filepath, lowest_only=False, lowest_n=False, energy_threshold=0.0
):
	out_molecules = []

	molecules = [mol for mol in Chem.SDMolSupplier(filepath, removeHs=False)]
	energies = [float(mol.GetProp("Energy")) for mol in molecules]
	min_energy = energies[0]
	for mol, energy in zip(molecules, energies):
		is_in_threshold = energy - min_energy < energy_threshold
		if lowest_n and is_in_threshold:
			out_molecules.append(mol)
		elif lowest_n:
			break
		elif lowest_only:
			out_molecules.append(mol)
			break
		else:
			out_molecules.append(mol)
	return out_molecules


def load_charge_data(filepath, backup_files):
	# read in dup_data to get the overall charge of MOLECULES
	invalid_files = []
	try:
		charge_data = pd.read_csv(filepath, usecols=["Molecule", "Overall charge"])
	except:
		charge_data = pd.DataFrame()
		for i, sdf_file in enumerate(backup_files):
			if not (Path(sdf_file).exists()):
				invalid_files.append(sdf_file)
				maxsplit = 1
				if "filter" in sdf_file:
					maxsplit += 1
				name = sdf_file.rsplit("_", maxsplit[0])
				charge = "Invalid"
			else:
				mol = next(pybel.readfile(sdf_file))
				name = mol.title.split(maxsplit=1)[0]
				charge = mol.data["Real charge"]
			charge_data.at[i, "Molecule"] = name
			charge_data.at[i, "Overall charge"] = charge
	return charge_data, invalid_files



try:
	import pybel
except ImportError:
	from openbabel import pybel # for openbabel>=3.0.0
from aqme.utils import periodic_table

def passes_custom_rules(mol,args,log):
	offset = args.angle_off
	for rule in args.geom_rules:
		passing = apply_rule(rule,mol,log,offset)
		if not passing:
			return False
	return True
def apply_rule(rule,mol,log,offset=0.0):
	if rule == 'Ir_bidentate_x3':
		return apply_Ir_rule(mol,offset)
	return apply_general_rule(rule,mol,log,offset)

def apply_general_rule(rule,mol,log,offset=0.0):
	try:
		syms = rule.split(',')[0].split('-')
		target_angle = float(rule.split(',')[1])
	except IndexError:
		log.write('x  The geom_rules parameter(s) was not correctly defined, this filter will be turned off')
		raise ValueError(f'rule "{rule}" is not properly defined')
	atnums = [periodic_table.index(sym) for sym in syms]
	smarts = pybel.Smarts('[#{}]~[#{}]~[#{}]'.format(*atnums))
	matches = smarts.findall(mol)
	if len(matches) >= 2:
		log.write(f'x  There are multiple options in geom_rules for {mol.title}, this filter will be turned off')
		log.write(f'x  {mol.title} contain more than one atom that meets the geom_rules criteria, this filter will be turned off')

	atoms = [mol.atoms[i].OBAtom for i in matches[0]]
	angle = mol.OBMol.GetAngle(*atoms)
	min_angle = target_angle - offset
	max_angle = target_angle + offset
	if min_angle <= angle <= max_angle:
		return False
	return True
def apply_Ir_rule(mol,offset=0.0):
	smarts = pybel.Smarts('[#77]1~[#7]~*~*~[#6]1')
	matches = smarts.findall(mol)
	min_angle = 180 - offset
	max_angle = 180 + offset
	if len(matches) == 3:
		# If 3 Py_Ph are coordinated Ns must not be at ~180º
		Ir = mol.atoms[matches[0][0]].OBAtom
		n1 = mol.atoms[matches[0][1]].OBAtom
		n2 = mol.atoms[matches[1][1]].OBAtom
		n3 = mol.atoms[matches[2][1]].OBAtom
		angles_atoms = [(n1,Ir,n2),
						(n1,Ir,n3),
						(n2,Ir,n3)]
		for i,j,k in angles_atoms:
			angle = mol.OBMol.GetAngle(i,j,k)
			if min_angle <= angle <= max_angle:
				return False

	elif len(matches) == 2:
		# If only 2 Py_Ph are coordinated Ns must be at ~180º
		Ir = mol.atoms[matches[0][0]].OBAtom
		n1 = mol.atoms[matches[0][1]].OBAtom
		n2 = mol.atoms[matches[1][1]].OBAtom
		angle = mol.OBMol.GetAngle(n1,Ir,n2)
		if not(min_angle <= angle <= max_angle):
			return False

	return True



#!/usr/bin/env python

#########################################################################################.
#########################################################################################
###                                                                                   ###
###  AQME is a tool that allows to carry out automated:                               ###
###  (1) Conformational searches and creation of COM files using RDKit, xTB and ANI   ###
###  (2) LOG file processing (detects imaginary freqs and error terminations          ###
###      and creates new COM files)                                                   ###
###  (3) Use LOG files to create new COM files with new keywords (i.e. single-point   ###
###      corrections after geometry optimization)                                     ###
###                                                                                   ###
#########################################################################################
###                                                                                   ###
###  Version: v1.0.1, Release date: 22-May-2020                                       ###
###                                                                                   ###
#########################################################################################
###                                                                                   ###
###  Authors: Shree Sowndarya S. V., Juan V. Alegre Requena, Robert S. Paton          ###
###                                                                                   ###
###  Please, report any bugs or suggestions to:                                       ###
###  svss@colostate.edu or juanvi89@hotmail.com                                       ###
###                                                                                   ###
#########################################################################################
#########################################################################################.

import os
import time
from pathlib import Path
from aqme.argument_parser import parser_args

from aqme.mainf import (
    csearch_main,
    geom_rules_main,
    qprep_main,
    graph_main,
    geom_par_main,
    nmr_main,
    energy_main,
    load_from_yaml,
    dbstep_par_main,
    nics_par_main,
    cclib_main,
    cmin_main,
)
from aqme.utils import Logger
from aqme.qcorr import qcorr, json2input


def main():
    # working directory and arguments
    w_dir_initial = os.getcwd()
    base_path = Path(w_dir_initial)

    args = parser_args()

    log_overall = Logger("aqme", args.output_name)
    # if needed to load from a yaml file
    args, log = load_from_yaml(args, log_overall)

    name = args.input.split(".")[0]

    # CSEARCH AND CMIN
    if args.CSEARCH in [
        "rdkit",
        "summ",
        "fullmonte",
        "crest",
    ]:  # RAUL: Is there any other posibilities or just None?
        start_time_overall = time.time()
        csearch_dup_data = csearch_main(w_dir_initial, args, log_overall)
        if args.time:
            elapsed_time = round(time.time() - start_time_overall, 2)
            log_overall.write(
                f"\n All molecules execution time CSEARCH: {elapsed_time} seconds"
            )
        os.chdir(w_dir_initial)
        if args.CMIN is None:
            csearch_csv_folder = base_path.joinpath("CSEARCH/csv_files")
            csearch_csv_folder.mkdir(exist_ok=True)
            csearch_csv_file = csearch_csv_folder.joinpath(f"{name}-CSEARCH-Data.csv")
            csearch_dup_data.to_csv(csearch_csv_file, index=False)

    # Separating CMIN
    if args.CSEARCH != None and args.CMIN in ["xtb", "ani"]:
        cmin_dup_data = cmin_main(w_dir_initial, args, log_overall, csearch_dup_data)
        if args.time:
            elapsed_time = round(time.time() - start_time_overall, 2)
            log_overall.write(
                f"\n All molecules execution time CMIN: {elapsed_time} seconds"
            )
        os.chdir(w_dir_initial)
        cmin_csv_folder = base_path.joinpath("CMIN/csv_files")
        cmin_csv_folder.mkdir(exist_ok=True)
        cmin_csv_file = cmin_csv_folder.joinpath(f"{name}-CMIN-Data.csv")
        cmin_dup_data.to_csv(cmin_csv_file, index=False)

    # applying rules to discard certain conformers based on rules that the user define
    if len(args.geom_rules) >= 1:
        geom_rules_active = True
        if args.qcorr == "gaussian":
            geom_rules_active = False
        geom_rules_main(args, log_overall, geom_rules_active)
        os.chdir(w_dir_initial)

    # QPREP
    if args.QPREP == "gaussian" or args.QPREP == "orca" or args.QPREP == "turbomole":
        qprep_main(w_dir_initial, args, log_overall)
        os.chdir(w_dir_initial)

    # if args.CSEARCH in ['rdkit', 'summ', 'fullmonte','crest'] or args.QPREP is not None:
    #     # moving files after compute and/or write_gauss
    #     move_sdf_main(args)
    #     os.chdir(w_dir_initial)


    #QCORR
    if args.qcorr:
        qcorr(qm_files=args.qm_files, w_dir_main=args.w_dir_main, dup_threshold=args.dup_threshold,
				mem=args.mem, nprocs=args.nprocs, chk=args.chk, qm_input=args.qm_input, 
                s2_threshold=args.s2_threshold, isom=args.isom,	isom_inputs=args.isom_inputs, 
                vdwfrac=args.vdwfrac, covfrac=args.covfrac, bs_gen=args.bs_gen, bs=args.bs, 
                gen_atoms=args.gen_atoms, qm_end=args.qm_end, amplitude_ifreq=args.amplitude_ifreq, 
                freq_conv=args.freq_conv, ifreq_cutoff=args.ifreq_cutoff, fullcheck=args.fullcheck,
                author=args.author, program=args.program, varfile=None)

    if args.json2input:
        json2input(json_files=args.json_files, w_dir_main=args.w_dir_main, destination=args.destination,
                suffix=args.suffix, charge=args.charge, mult=args.mult,	mem=args.mem, nprocs=args.nprocs,
                chk=args.chk, qm_input=args.qm_input, bs_gen=args.bs_gen, bs=args.bs, gen_atoms=args.gen_atoms,
                qm_end=args.qm_end, program=args.program)

    # QPRED
    if args.QPRED == "nmr":
        nmr_main(args, log_overall, w_dir_initial)
    if args.QPRED == "energy":
        energy_main(args, log_overall, w_dir_initial)
    if args.QPRED == "dbstep":
        dbstep_par_main(args, log_overall, w_dir_initial)
    if args.QPRED == "nics":
        nics_par_main(args, log_overall, w_dir_initial)
    if args.QPRED == "cclib-json":
        cclib_main(args, log_overall, w_dir_initial)
    os.chdir(w_dir_initial)

    # QSTAT
    if args.QSTAT == "descp":
        geom_par_main(args, log_overall, w_dir_initial)
    if args.QSTAT == "graph":
        graph_main(args, log_overall, w_dir_initial)
    os.chdir(w_dir_initial)

    log_overall.finalize()

    out_data_file = Path(f"aqme_output.dat")
    if out_data_file.exists():
        out_data_file.replace(f"aqme_{args.output_name}.dat")


if __name__ == "__main__":
    main()



#####################################################.
#       This file stores all the main functions     #
#####################################################.

import glob
import sys
import os
import concurrent.futures as futures
import multiprocessing as mp
from pathlib import Path

import pandas as pd
from progress.bar import IncrementalBar
from rdkit.Chem import AllChem as Chem

# from aqme.csearch import (check_charge_smi, clean_args,
#                                compute_confs,
#                                mol_from_sdf_or_mol_or_mol2, creation_of_dup_csv)
from aqme.csearch import *
from aqme.filter import geom_rules_output

from aqme.qprep import qprep, get_molecule_list, load_charge_data

from aqme.cmin import mult_min
from aqme.utils import (
	move_file,
	get_filenames,
	creation_of_dup_csv_csearch,
	creation_of_dup_csv_cmin,
)

# need to and in energy


def csearch_main(w_dir_initial, args, log_overall):

	file_format = os.path.splitext(args.input)[1]

	# Checks
	if file_format not in SUPPORTED_INPUTS:
		log_overall.write("\nx  INPUT FILETYPE NOT CURRENTLY SUPPORTED!")
		sys.exit()
	if not os.path.exists(args.input):
		log_overall.write("\nx  INPUT FILE NOT FOUND!")
		sys.exit()

	# if large system increase stack size
	# if args.STACKSIZE != '1G':
	#     os.environ['OMP_STACKSIZE'] = args.STACKSIZE
	smi_derivatives = [".smi", ".txt", ".yaml", ".yml", ".rtf"]
	Extension2inputgen = dict()
	for key in smi_derivatives:
		Extension2inputgen[key] = prepare_smiles_files
	Extension2inputgen[".csv"] = prepare_csv_files
	Extension2inputgen[".cdx"] = prepare_cdx_files
	Extension2inputgen[".gjf"] = prepare_gaussian_files
	Extension2inputgen[".com"] = prepare_gaussian_files
	Extension2inputgen[".xyz"] = prepare_gaussian_files
	Extension2inputgen[".sdf"] = prepare_sdf_files
	Extension2inputgen[".mol"] = prepare_mol_files
	Extension2inputgen[".mol2"] = prepare_mol_files

	with futures.ProcessPoolExecutor(
		max_workers=args.cpus, mp_context=mp.get_context("fork")
	) as executor:
		# Submit a set of asynchronous jobs
		jobs = []
		count_mol = 0

		# Prepare the Jobs
		prepare_function = Extension2inputgen[file_format]
		job_inputs = prepare_function(args, w_dir_initial)

		# Submit the Jobs
		for job_input in job_inputs:
			smi_, name_, dir_, varfile_, charge_default_, constraints_dist_, constraints_angle_, constraints_dihedral_ = job_input
			job = executor.submit(
				process_csearch, smi_, name_, dir_, varfile_, charge_default_, constraints_dist_, constraints_angle_, constraints_dihedral_
			)
			jobs.append(job)
			count_mol += 1

		final_dup_data = creation_of_dup_csv_csearch(args.CSEARCH)
		bar = IncrementalBar("o  Number of finished jobs from CSEARCH", max=count_mol)
		# Process the job results (in submission order) and save the conformers.
		for i, job in enumerate(jobs):
			total_data = job.result()
			frames = [final_dup_data, total_data]
			final_dup_data = pd.concat(frames, ignore_index=True, sort=True)
			bar.next()
		bar.finish()

		# removing temporary files
		temp_files = [
			"gfn2.out",
			"xTB_opt.traj",
			"ANI1_opt.traj",
			"wbo",
			"xtbrestart",
			"ase.opt",
			"xtb.opt",
			"gfnff_topo",
		]
		for file in temp_files:
			if os.path.exists(file):
				os.remove(file)

	return final_dup_data


def cmin_main(w_dir_initial, args, log_overall, dup_data):
	bar = IncrementalBar("o  Number of finished jobs from CMIN", max=len(dup_data))
	final_dup_data = creation_of_dup_csv_cmin(args.CMIN)
	for dup_data_idx in range(len(dup_data)):
		# update_to_rdkit = dup_data.at[dup_data_idx,'update_to_rdkit']
		name = dup_data.at[dup_data_idx, "Molecule"]
		charge = dup_data.at[dup_data_idx, "Overall charge"]
		if dup_data.at[dup_data_idx, "status"] != -1:
			if args.CMIN == "ani":
				min_suffix = "ani"
			elif args.CMIN == "xtb":
				min_suffix = "xtb"
			if args.CSEARCH in ["rdkit", "summ", "fullmonte"]:

				csearch_folder = Path(w_dir_initial).joinpath(f"CSEARCH/{args.CSEARCH}")
				fullname = str(csearch_folder.joinpath(name + "_" + args.CSEARCH))

				# fullname = f'{name}_{args.CSEARCH}'
				# try:
				total_data = mult_min(
					fullname, args, min_suffix, charge, log_overall, w_dir_initial
				)
				# except:
				#     pass
				frames = [final_dup_data, total_data]
				final_dup_data = pd.concat(frames, ignore_index=True, sort=True)
		bar.next()
	bar.finish()

	# removing temporary files
	temp_files = [
		"gfn2.out",
		"xTB_opt.traj",
		"ANI1_opt.traj",
		"wbo",
		"xtbrestart",
		"ase.opt",
		"xtb.opt",
		"gfnff_topo",
	]
	for file in temp_files:
		if Path(file).exists():
			os.remove(file)

	return final_dup_data


# MAIN QPREP FUNCTION
def qprep_main(w_dir_initial, args, log):

	if len(args.geom_rules) >= 1:
		conf_files = glob.glob("*_rules.sdf")
	# define the SDF files to convert to COM Gaussian files
	elif args.CMIN == "xtb":
		conf_files = glob.glob(w_dir_initial + "/CMIN/" + args.CMIN + "/*_xtb.sdf")
	elif args.CMIN == "ani":
		conf_files = glob.glob(w_dir_initial + "/CMIN/" + args.CMIN + "/*_ani.sdf")
	elif args.CSEARCH == "rdkit":
		conf_files = glob.glob(
			w_dir_initial + "/CSEARCH/" + args.CSEARCH + "/*_rdkit.sdf"
		)
	elif args.CSEARCH == "summ":
		conf_files = glob.glob(
			w_dir_initial + "/CSEARCH/" + args.CSEARCH + "/*_summ.sdf"
		)
	elif args.CSEARCH == "fullmonte":
		conf_files = glob.glob(
			w_dir_initial + "/CSEARCH/" + args.CSEARCH + "/*_fullmonte.sdf"
		)
	elif args.CSEARCH == "crest":
		conf_files = glob.glob(
			w_dir_initial + "/CSEARCH/" + args.CSEARCH + "/*_crest.sdf"
		)
	else:
		conf_files = glob.glob("*.sdf")

	# # NEED TO UPDATE THIS PART TO START FROM JSON!
	# if args.com_from_xyz:
	#     xyz_files = glob.glob("*.xyz")
	#     for file in xyz_files:
	#         mol = next(pybel.readfile("xyz", file))
	#         stem = Path(file).stem
	#         mol.write("sdf", f"{stem}.sdf")
	#     conf_files = glob.glob("*.sdf")

	if not conf_files:
		log.write("\nx  No SDF files detected to convert to gaussian COM files")
		return

	csv_name = args.input.split(".")[0]
	csv_file = f"{w_dir_initial}/CSEARCH/csv_files/{csv_name}-CSEARCH-Data.csv"
	charge_data, invalid_files = load_charge_data(csv_file, conf_files)

	# remove the invalid files and non-existing files
	accept_file = lambda x: x not in invalid_files and Path(x).exists()
	conf_files = [file for file in conf_files if accept_file(file)]

	# Prepare the list of molecules that are to be written
	mols = []

	for file in conf_files:
		filepath = f"{file}"
		new_mols = get_molecule_list(
			filepath,
			lowest_only=args.lowest_only,
			lowest_n=args.lowest_n,
			energy_threshold=args.energy_threshold_for_gaussian,
		)
		mols.extend(new_mols)

		name = os.path.basename(filepath).split(".")[0].split('_')[0]
		charge = charge_data[charge_data.Molecule == name]['Overall charge'][0]

		# writing the com files
		for i, mol in enumerate(mols):
			qprep(mol=mol, molecule=name+'_conf_'+str(i+1), charge=charge, atom_types = [], yaml_file=args.varfile)

      
# moving files after compute and/or write_gauss
def move_sdf_main(args):
	src = Path(os.getcwd())
	if len(args.geom_rules) >= 1:
		geom_rules_files = glob.glob("*_filter_geom_rules.sdf")
	if args.CMIN == "xtb":
		all_xtb_conf_files = glob.glob("*_xtb.sdf")
		destination_xtb = src.joinpath("CMIN/xtb/")
		for file in all_xtb_conf_files:
			move_file(destination_xtb, src, file)
		all_xtb_conf_files_all = glob.glob("*_xtb_all_confs.sdf")
		destination_xtb_all = src.joinpath("CMIN/xtb_all_confs/")
		for file in all_xtb_conf_files_all:
			move_file(destination_xtb_all, src, file)
		if len(args.geom_rules) >= 1:
			destination_geom_rules = src.joinpath("CMIN/xtb/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)
	if args.CMIN == "ani":
		all_ani_conf_files = glob.glob("*_ani.sdf")
		destination_ani = src.joinpath("CMIN/ani/")
		for file in all_ani_conf_files:
			move_file(destination_ani, src, file)
		all_ani_conf_files_all = glob.glob("*_ani_all_confs.sdf")
		destination_ani_all = src.joinpath("CMIN/ani_all_confs/")
		for file in all_ani_conf_files_all:
			move_file(destination_ani_all, src, file)
		if len(args.geom_rules) >= 1:
			destination_geom_rules = src.joinpath("CMIN/ani/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)
	if args.CSEARCH == "rdkit":
		all_name_conf_files = glob.glob("*_rdkit.sdf")
		destination_rdkit = src.joinpath("CSEARCH/rdkit/")
		for file in all_name_conf_files:
			move_file(destination_rdkit, src, file)
		if len(args.geom_rules) >= 1:
			destination_geom_rules = src.joinpath("CSEARCH/rdkit/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)

	if args.CSEARCH == "summ":
		all_name_conf_files = glob.glob("*_summ.sdf")
		destination_rdkit = src.joinpath("CSEARCH/summ/")
		for file in all_name_conf_files:
			move_file(destination_rdkit, src, file)
		if len(args.geom_rules) >= 1 and args.CMIN is None:
			destination_geom_rules = src.joinpath("CSEARCH/summ/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)

	if args.CSEARCH == "fullmonte":
		all_name_conf_files = glob.glob("*_fullmonte.sdf")
		destination_rdkit = src.joinpath("CSEARCH/fullmonte/")
		for file in all_name_conf_files:
			move_file(destination_rdkit, src, file)
		if len(args.geom_rules) >= 1 and args.CMIN is None:
			destination_geom_rules = src.joinpath("CSEARCH/fullmonte/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)

	if args.CSEARCH == "crest":
		all_name_conf_files = glob.glob("*_crest.sdf")
		destination_rdkit = src.joinpath("CSEARCH/crest/")
		for file in all_name_conf_files:
			move_file(destination_rdkit, src, file)
		if len(args.geom_rules) >= 1 and args.CMIN is None:
			destination_geom_rules = src.joinpath("CSEARCH/crest/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)

	if args.CSEARCH is None:
		if len(args.geom_rules) >= 1:
			destination_geom_rules = src.joinpath("QCALC/SDF_input/filter_geom_rules/")
			for file in geom_rules_files:
				move_file(destination_geom_rules, src, file)

		all_conf_files = glob.glob("*.sdf")
		destination = src.joinpath("QCALC/SDF_input/")
		for file in all_conf_files:
			move_file(destination, src, file)

	if args.com_from_xyz:
		all_xyz_conf_files = glob.glob("*.xyz") + glob.glob("*.sdf")
		destination_xyz = src.joinpath("QCALC/xyz_and_sdf/")
		for file in all_xyz_conf_files:
			move_file(destination_xyz, src, file)


# getting descriptors
def geom_par_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no  Calculating paramters for molecule : {0} ".format(name))

		sdf_ani, sdf_xtb = None, None
		if os.path.exists(w_dir_initial + "/CSEARCH/rdkit/" + name + "_rdkit.sdf"):
			sdf_rdkit = w_dir_initial + "/CSEARCH/rdkit/" + name + "_rdkit.sdf"
		elif os.path.exists(w_dir_initial + "/CSEARCH/summ/" + name + "_summ.sdf"):
			sdf_rdkit = w_dir_initial + "/CSEARCH/summ/" + name + "_summ.sdf"
		elif os.path.exists(
			w_dir_initial + "/CSEARCH/fullmonte/" + name + "_fullmonte.sdf"
		):
			sdf_rdkit = w_dir_initial + "/CSEARCH/fullmonte/" + name + "_fullmonte.sdf"
		if os.path.exists(w_dir_initial + "/CMIN/xtb/" + name + "_xtb.sdf"):
			sdf_xtb = w_dir_initial + "/CMIN/xtb/" + name + "_xtb.sdf"
		if os.path.exists(w_dir_initial + "/CMIN/ani/" + name + "_ani.sdf"):
			sdf_ani = w_dir_initial + "/CMIN/ani/" + name + "_ani.sdf"
		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
			# Sets the folder and find the log files to analyze
			for lot, bs, bs_gcp in zip(
				args.level_of_theory, args.basis_set, args.genecp_bs
			):
				# assign the path to the finished directory.
				if str(bs).find("/") > -1:
					w_dir = (
						args.path
						+ str(lot)
						+ "-"
						+ str(bs).split("/")[0]
						+ "/success/output_files"
					)
				else:
					w_dir = (
						args.path + str(lot) + "-" + str(bs) + "/success/output_files"
					)
				os.chdir(w_dir)
				qm_files = get_filenames("output", name)
				calculate_parameters(
					sdf_rdkit,
					sdf_ani,
					sdf_xtb,
					qm_files,
					args,
					log,
					w_dir_initial,
					name,
					lot,
					bs,
				)
		else:
			calculate_parameters(
				sdf_rdkit,
				sdf_ani,
				sdf_xtb,
				None,
				args,
				log,
				w_dir_initial,
				name,
				None,
				None,
			)

		os.chdir(w_dir_initial)


# function to plot graphs
def graph_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no  Plotting graphs for molecule : {0} ".format(name))

		sdf_ani, sdf_xtb = None, None
		if os.path.exists(w_dir_initial + "/CSEARCH/rdkit/" + name + "_rdkit.sdf"):
			sdf_rdkit = w_dir_initial + "/CSEARCH/rdkit/" + name + "_rdkit.sdf"
		elif os.path.exists(w_dir_initial + "/CSEARCH/summ/" + name + "_summ.sdf"):
			sdf_rdkit = w_dir_initial + "/CSEARCH/summ/" + name + "_summ.sdf"
		elif os.path.exists(
			w_dir_initial + "/CSEARCH/fullmonte/" + name + "_fullmonte.sdf"
		):
			sdf_rdkit = w_dir_initial + "/CSEARCH/fullmonte/" + name + "_fullmonte.sdf"
		if os.path.exists(
			w_dir_initial + "/CMIN/xtb_all_confs/" + name + "_xtb_all_confs.sdf"
		):
			sdf_xtb = (
				w_dir_initial + "/CMIN/xtb_all_confs/" + name + "_xtb_all_confs.sdf"
			)
		if os.path.exists(
			w_dir_initial + "/CMIN/ani_all_confs/" + name + "_ani_all_confs.sdf"
		):
			sdf_ani = (
				w_dir_initial + "/CMIN/ani_all_confs/" + name + "_ani_all_confs.sdf"
			)
		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
			# Sets the folder and find the log files to analyze
			for lot, bs, bs_gcp in zip(
				args.level_of_theory, args.basis_set, args.genecp_bs
			):
				# assign the path to the finished directory.
				w_dir = args.path + str(lot) + "-" + str(bs) + "/success/output_files"
				os.chdir(w_dir)
				qm_files = get_filenames("output", name)
				if os.path.exists(
					args.path + str(lot) + "-" + str(bs) + "/success/G16-SP_input_files"
				):
					for lot_sp, bs_sp, bs_gcp_sp in zip(
						args.level_of_theory_sp, args.basis_set_sp, args.gen_bs_sp
					):
						w_dir_sp = (
							args.path
							+ str(lot)
							+ "-"
							+ str(bs)
							+ "/success/G16-SP_input_files"
							+ "/"
							+ str(lot_sp)
							+ "-"
							+ str(bs_sp)
						)
						sp_files = get_filenames("output", name)
						graph(
							sdf_rdkit,
							sdf_xtb,
							sdf_ani,
							qm_files,
							sp_files,
							args,
							log,
							lot,
							bs,
							lot_sp,
							bs_sp,
							name,
							w_dir_initial,
							w_dir_sp,
							w_dir,
							"g16",
						)
				if os.path.exists(
					f"{args.path}{lot}-{bs}/success/TURBOMOLE-SP_input_files"
				):
					for lot_sp, bs_sp, bs_gcp_sp in zip(
						args.level_of_theory_sp, args.basis_set_sp, args.gen_bs_sp
					):
						w_dir_sp = f"{path_turbomole}/{lot_sp}-{bs_sp.split('/')[0]}"
						os.chdir(w_dir_sp)
						sp_files = []
						for path in Path(w_dir_sp).iterdir():
							if path.is_dir() and "_SP" in path.stem:
								sp_files.append(path)
						os.chdir(w_dir)
						graph(
							sdf_rdkit,
							sdf_xtb,
							sdf_ani,
							qm_files,
							sp_files,
							args,
							log,
							lot,
							bs,
							lot_sp,
							bs_sp.split("/")[0],
							name,
							w_dir_initial,
							w_dir_sp,
							w_dir,
							"turbomole",
						)

				if os.path.exists(
					args.path
					+ str(lot)
					+ "-"
					+ str(bs)
					+ "/success/ORCA-SP_input_files"
				):
					for lot_sp, bs_sp, bs_gcp_sp in zip(
						args.level_of_theory_sp, args.basis_set_sp, args.gen_bs_sp
					):
						w_dir_sp = (
							args.path
							+ str(lot)
							+ "-"
							+ str(bs)
							+ "/success/ORCA-SP_input_files"
							+ "/"
							+ str(lot_sp)
							+ "-"
							+ str(bs_sp.split("/")[0])
						)
						os.chdir(w_dir_sp)
						sp_files = get_filenames("output", name)
						os.chdir(w_dir)
						graph(
							sdf_rdkit,
							sdf_xtb,
							sdf_ani,
							qm_files,
							sp_files,
							args,
							log,
							lot,
							bs,
							lot_sp,
							bs_sp.split("/")[0],
							name,
							w_dir_initial,
							w_dir_sp,
							w_dir,
							"orca",
						)
				else:
					graph(
						sdf_rdkit,
						sdf_xtb,
						sdf_ani,
						qm_files,
						None,
						args,
						log,
						lot,
						bs,
						None,
						None,
						name,
						w_dir_initial,
						None,
						w_dir,
						None,
					)

		else:
			graph(
				sdf_rdkit,
				sdf_xtb,
				sdf_ani,
				None,
				None,
				args,
				log,
				None,
				None,
				None,
				None,
				name,
				w_dir_initial,
				None,
				None,
				None,
			)

	os.chdir(w_dir_initial)


# function for comparison of nmr
def nmr_main(args, log, w_dir_initial):

	if os.path.exists(w_dir_initial + "/QCALC/G16"):
		args.path = w_dir_initial + "/QCALC/G16/"
	else:
		log.write(
			"\nx  The path for NMR analysis was not set up properly! (check the tutorials for more information)"
		)

	# get sdf FILES from csv
	try:
		pd_name = pd.read_csv(
			w_dir_initial
			+ "/CSEARCH/csv_files/"
			+ args.input.split(".")[0]
			+ "-CSEARCH-Data.csv"
		)

	except FileNotFoundError:
		# detects all the unique molecules from the success folder
		if str(args.basis_set[0]).find("/") > -1:
			w_dir_fin = (
				args.path
				+ str(args.level_of_theory[0])
				+ "-"
				+ str(args.basis_set[0]).split("/")[0]
				+ "/success/output_files"
			)
		else:
			w_dir_fin = (
				args.path
				+ str(args.level_of_theory[0])
				+ "-"
				+ str(args.basis_set[0])
				+ "/success/output_files"
			)
		os.chdir(w_dir_fin)

		nmr_list = []
		standard_suffixes = ["xtb", "ani"]
		for name_nmr in glob.glob("*.*"):
			# discard_charact keeps track of the extra characters after the name of the molecule
			discard_charact = len(name_nmr.split(".")[1]) + 1
			potential_unique = name_nmr.split(".")[0].split("_")
			for i in reversed(range(len(potential_unique))):
				try:
					if potential_unique[i] not in standard_suffixes:
						int(potential_unique[i])
					discard_charact += len(potential_unique[i])
					discard_charact += 1
				except ValueError:
					original_name = name_nmr[:-discard_charact]
					if original_name not in nmr_list:
						nmr_list.append(original_name)
		pd_name = pd.DataFrame(data=nmr_list, columns=["Molecule"])

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no NMR analysis for molecule : {0} ".format(name))

		# Sets the folder and find the log files to analyze
		for lot, bs, bs_gcp in zip(
			args.level_of_theory, args.basis_set, args.genecp_bs
		):
			# assign the path to the finished directory.
			if str(bs).find("/") > -1:
				w_dir_fin = (
					args.path
					+ str(lot)
					+ "-"
					+ str(bs).split("/")[0]
					+ "/success/output_files"
				)
			else:
				w_dir_fin = (
					args.path + str(lot) + "-" + str(bs) + "/success/output_files"
				)
			os.chdir(w_dir_fin)

			qm_files = get_filenames("output", name)
			if len(qm_files) != 0:
				calculate_boltz_and_nmr(
					qm_files, args, log, name, w_dir_fin, w_dir_initial, lot, bs
				)
	os.chdir(w_dir_initial)


def energy_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write(
			"\no Boltzmann average energy analysis for molecule : {0} ".format(name)
		)
		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
		# Sets the folder and find the log files to analyze
		for lot, bs, bs_gcp in zip(
			args.level_of_theory, args.basis_set, args.genecp_bs
		):
			# assign the path to the finished directory.
			if str(bs).find("/") > -1:
				w_dir_fin = (
					args.path
					+ str(lot)
					+ "-"
					+ str(bs).split("/")[0]
					+ "/success/output_files/"
				)
			else:
				w_dir_fin = (
					args.path + str(lot) + "-" + str(bs) + "/success/output_files/"
				)
			os.chdir(w_dir_fin)
			qm_files = get_filenames("output", name)
			if len(qm_files) != 0:
				calculate_boltz_and_energy(
					qm_files, args, log, name, w_dir_fin, w_dir_initial, lot, bs
				)

	# combining the combining all files in different folders
	w_dir_boltz = w_dir_initial + "/QPRED/energy/boltz/"

	for lot, bs, bs_gcp in zip(args.level_of_theory, args.basis_set, args.genecp_bs):
		# assign the path to the finished directory.
		if str(bs).find("/") > -1:
			w_dir_fin = w_dir_boltz + str(lot) + "-" + str(bs).split("/")[0]
		else:
			w_dir_fin = w_dir_boltz + str(lot) + "-" + str(bs)
		os.chdir(w_dir_fin)
		dat_files = glob.glob("*.dat")
		if len(dat_files) != 0:
			calculate_avg_and_energy(
				dat_files,
				args,
				log,
				name,
				w_dir_fin,
				w_dir_initial,
				w_dir_boltz,
				lot,
				bs,
			)

	os.chdir(w_dir_initial)


def dbstep_par_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no  Calculating paramters for molecule : {0} ".format(name))

		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
			# Sets the folder and find the log files to analyze
			for lot, bs, bs_gcp in zip(
				args.level_of_theory, args.basis_set, args.genecp_bs
			):
				# assign the path to the finished directory.
				if str(bs).find("/") > -1:
					w_dir = (
						args.path
						+ str(lot)
						+ "-"
						+ str(bs).split("/")[0]
						+ "/success/output_files"
					)
				else:
					w_dir = (
						args.path + str(lot) + "-" + str(bs) + "/success/output_files"
					)
				os.chdir(w_dir)
				qm_files = get_filenames("output", name)
				calculate_db_parameters(
					qm_files, args, log, w_dir_initial, name, lot, bs
				)
				calculate_boltz_and_dbstep(
					qm_files, args, log, name, w_dir, w_dir_initial, lot, bs
				)
		os.chdir(w_dir_initial)


def nics_par_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no  Calculating nics for molecule : {0} ".format(name))

		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
			# Sets the folder and find the log files to analyze
			for lot, bs, bs_gcp in zip(
				args.level_of_theory, args.basis_set, args.genecp_bs
			):
				# assign the path to the finished directory.
				if str(bs).find("/") > -1:
					w_dir = (
						args.path
						+ str(lot)
						+ "-"
						+ str(bs).split("/")[0]
						+ "/success/output_files"
					)
				else:
					w_dir = (
						args.path + str(lot) + "-" + str(bs) + "/success/output_files"
					)
				os.chdir(w_dir)
				qm_files = get_filenames("output", name)
				# do boltz firsst
				calculate_boltz_for_nics(
					qm_files, args, log, name, w_dir, w_dir_initial, lot, bs
				)
				for lot_sp, bs_sp, bs_gcp_sp in zip(
					args.level_of_theory_sp, args.basis_set_sp, args.gen_bs_sp
				):
					if str(bs).find("/") > -1:
						w_dir_sp = (
							args.path
							+ str(lot)
							+ "-"
							+ str(bs).split("/")[0]
							+ "/success/G16-NICS_input_files/"
							+ str(lot_sp)
							+ "-"
							+ str(bs_sp)
						)
					else:
						w_dir_sp = (
							args.path
							+ str(lot)
							+ "-"
							+ str(bs)
							+ "/success/G16-NICS_input_files/"
							+ str(lot_sp)
							+ "-"
							+ str(bs_sp)
						)
					os.chdir(w_dir_sp)
					qm_files_sp = get_filenames("output", name)
					calculate_nics_parameters(
						qm_files_sp, args, log, w_dir_initial, name, lot_sp, bs_sp
					)
					calculate_avg_nics(
						qm_files_sp,
						args,
						log,
						name,
						w_dir_sp,
						w_dir_initial,
						lot_sp,
						bs_sp,
					)
		os.chdir(w_dir_initial)


def cclib_main(args, log, w_dir_initial):
	# get sdf FILES from csv
	pd_name = pd.read_csv(
		w_dir_initial
		+ "/CSEARCH/csv_files/"
		+ args.input.split(".")[0]
		+ "-CSEARCH-Data.csv"
	)

	for i in range(len(pd_name)):
		name = pd_name.loc[i, "Molecule"]

		log.write("\no  Calculating cclib paramters for molecule : {0} ".format(name))
		if os.path.exists(w_dir_initial + "/QCALC/G16"):
			args.path = w_dir_initial + "/QCALC/G16/"
			# Sets the folder and find the log files to analyze
			for lot, bs, bs_gcp in zip(
				args.level_of_theory, args.basis_set, args.genecp_bs
			):
				# assign the path to the finished directory.
				if str(bs).find("/") > -1:
					w_dir = (
						args.path
						+ str(lot)
						+ "-"
						+ str(bs).split("/")[0]
						+ "/success/output_files"
					)
				else:
					w_dir = (
						args.path + str(lot) + "-" + str(bs) + "/success/output_files"
					)
				os.chdir(w_dir)
				qm_files = get_filenames("output", name)
				# do boltz firsst
				calculate_cclib(qm_files, w_dir_initial, lot, bs)
				calculate_boltz_for_cclib(qm_files, name, w_dir_initial, lot, bs)
				if str(bs).find("/") > -1:
					os.chdir(
						w_dir_initial
						+ "/QPRED/cclib-json/all_confs_cclib/"
						+ str(lot)
						+ "-"
						+ str(bs).split("/")[0]
					)
				else:
					os.chdir(
						w_dir_initial
						+ "/QPRED/cclib-json/all_confs_cclib/"
						+ str(lot)
						+ "-"
						+ str(bs)
					)
				json_files = get_filenames("output", name)
				get_avg_cclib_param(json_files, name, w_dir_initial, lot, bs)


# MAIN OPTION FOR DISCARDING MOLECULES BASED ON USER INPUT DATA (REFERRED AS EXPERIMENTAL RULES)
def geom_rules_main(args, log, geom_rules_active):
	if geom_rules_active:
		if args.verbose:
			log.write(
				"\n   ----- Applying experimental rules to write the new confs file -----"
			)
		# do 2 cases, for RDKit only and RDKIt+xTB
		if args.CMIN == "xtb":
			conf_files = glob.glob("*_xtb.sdf")
		if args.CMIN == "ani":
			conf_files = glob.glob("*_ani.sdf")
		if args.CMIN is None:
			if args.CSEARCH == "rdkit":
				conf_files = glob.glob("*_rdkit.sdf")
			elif args.CSEARCH == "summ":
				conf_files = glob.glob("*_summ.sdf")
			elif args.CSEARCH == "fullmonte":
				conf_files = glob.glob("*_fullmonte.sdf")
			else:
				conf_files = glob.glob("*.sdf")

		for file in conf_files:
			try:
				allmols = Chem.SDMolSupplier(file, removeHs=False)
			except OSError:
				pass
			if allmols:
				sdwriter = Chem.SDWriter(file.split(".")[0] + "_filter_geom_rules.sdf")
				print_error_geom_rules = False
				for mol in allmols:
					check_mol = True
					ob_compat = True
					rdkit_compat = True
					check_mol = geom_rules_output(
						mol, args, log, file, print_error_geom_rules
					)
					print_error_geom_rules += 1
					if check_mol:
						sdwriter.write(mol)

				sdwriter.close()



#####################################################.
#        This file stores all the functions         #
#           used for genrating a graph              #
#####################################################.

import os
from rdkit.Chem import AllChem as Chem
import numpy as np
from sklearn.metrics import mean_absolute_error
import statistics as stats

ev_2_kcal_mol = 23.061 #ev to kcal/mol
hartree_to_kcal = 627.509

def stats_calc(y_dft,y):
    y_diff = abs(np.subtract(y, y_dft))
    mae = mean_absolute_error(y, y_dft)
    sd = stats.stdev(y_diff)
    return mae,sd

def get_energy(inmols_min):
    energy_min = []
    for i,mol in enumerate(inmols_min):
        energy_min.append(["_".join(mol.GetProp('_Name').split(' ')),mol.GetProp('Energy')])
    return energy_min

def rename_name(energy,type):
    for i,_ in enumerate(energy):
        energy[i][0] = energy[i][0].split('_'+type)[0]
    return energy

def scaling_with_lowest(energy):
    #scaling arrays
    v = np.array(energy)[:, 1].astype(np.float)
    energy_sc = (v - v.min())
    for i,_ in enumerate(energy):
        energy[i][1] = energy_sc[i]
    return energy

def plot_graph(energy_rdkit,energy_min,energy_min_dft,lot,bs,energy_min_dft_sp,lot_sp,bs_sp,name_mol,args,log,type_csearch,type_min,w_dir_initial):

    try:
        import matplotlib.path as mpath
        import matplotlib.patches as mpatches
        import matplotlib.pyplot as plt
    except (ModuleNotFoundError,AttributeError):
        log.write('The Matplotlib module is not installed correctly - the graphing functions are not available')

    def get_cmap(n, name='viridis'):
        '''Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
        RGB color; the keyword argument name must be a standard mpl colormap name.'''
        return plt.cm.get_cmap(name, n)

    if len(energy_min_dft_sp)  != 0:
        energy_dft_sp_mae_sd,energy_dft_mae_sd,energy_min_mae_sd,energy_rdkit_mae_sd = [],[],[],[]
        for l,_ in enumerate(energy_min_dft_sp):
            name = energy_min_dft_sp[l][0]
            energy_dft_sp_mae_sd.append(float(energy_min_dft_sp[l][1]))
            for i,_ in enumerate(energy_min_dft):
                if energy_min_dft[i][0] == name:
                    energy_dft_mae_sd.append(float(energy_min_dft[i][1]))
            if energy_min is not None:
                for j,_ in enumerate(energy_min):
                    if energy_min[j][0] == name:
                        energy_min_mae_sd.append(float(energy_min[j][1]))
            if energy_rdkit is not None:
                for k,_ in enumerate(energy_rdkit):
                    if energy_rdkit[k][0] == name:
                        energy_rdkit_mae_sd.append(float(energy_rdkit[k][1]))

        mae_rdkit,sd_rdkit = stats_calc(energy_dft_sp_mae_sd,energy_rdkit_mae_sd)
        mae_min,sd_min = stats_calc(energy_dft_sp_mae_sd,energy_min_mae_sd)
        mae_dft,sd_dft = stats_calc(energy_dft_sp_mae_sd,energy_dft_mae_sd)

    elif len(energy_min_dft) != 0:
        energy_dft_mae_sd,energy_min_mae_sd,energy_rdkit_mae_sd = [],[],[]
        for i,_ in enumerate(energy_min_dft):
            name = energy_min_dft[i][0]
            energy_dft_mae_sd.append(float(energy_min_dft[i][1]))
            if energy_min is not None:
                for j,_ in enumerate(energy_min):
                    if energy_min[j][0] == name:
                        energy_min_mae_sd.append(float(energy_min[j][1]))
            if energy_rdkit is not None:
                for k,_ in enumerate(energy_rdkit):
                    if energy_rdkit[k][0] == name:
                        energy_rdkit_mae_sd.append(float(energy_rdkit[k][1]))

        mae_rdkit,sd_rdkit = stats_calc(energy_dft_mae_sd,energy_rdkit_mae_sd)
        mae_min,sd_min = stats_calc(energy_dft_mae_sd,energy_min_mae_sd)


    x_axis_names = []
    x_axis_names.append(type_csearch)
    if type_min is not None:
        x_axis_names.append(type_min)
    if lot is not None:
        x_axis_names.append(lot+'\n'+bs)
    if lot_sp is not None:
        x_axis_names.append(lot_sp+'\n'+bs_sp)

    x_axis_names_mae = []
    x_axis_names_mae.append(type_csearch)
    if type_min is not None:
        x_axis_names_mae.append(type_min)
    if lot is not None:
        x_axis_names_mae.append(lot+'-'+bs)
    if lot_sp is not None:
        x_axis_names_mae.append(lot_sp+'-'+bs_sp)

    x_axis = [0,1,2,3]
    x_axis_1 = [0,1,2]
    x_axis_2 = [0,1]
    x_axis_3 = [0]


    list_all = []
    name_all = []
    for i,_ in enumerate(energy_rdkit):
        list = []
        name = energy_rdkit[i][0]
        name_all.append(name)
        list.append(energy_rdkit[i][1])
        if energy_min is not None:
            for j,_ in enumerate(energy_min):
                if energy_min[j][0] == name:
                    list.append(energy_min[j][1])
        if energy_min_dft is not None:
            for k,_ in enumerate(energy_min_dft):
                if energy_min_dft[k][0] == name:
                    list.append(energy_min_dft[k][1])
        if energy_min_dft_sp is not None:
            for l,_ in enumerate(energy_min_dft_sp):
                if energy_min_dft_sp[l][0] == name:
                    list.append(energy_min_dft_sp[l][1])
        list_all.append(list)

    fig=plt.figure() #Creates a new figure
    ax1=fig.add_subplot(111) #Plot with: 1 row, 1 column, first subplot.

    cmap = get_cmap(len(list_all))
    Path = mpath.Path
    for i, list in enumerate(list_all):
        if len(list) == 2 or len(list) == 3 or len(list) == 4:
            path_patch_1 = mpatches.PathPatch(
                    Path([(x_axis[0], list[0]), (x_axis[0]+0.5, list[1]), (x_axis[1] ,list[1])],
                         [Path.MOVETO, Path.CURVE3, Path.CURVE3]),
                    fc="none", transform=ax1.transData, color=cmap(i))
            ax1.add_patch(path_patch_1)
        if len(list) == 3 or len(list) == 4:
            path_patch_2 = mpatches.PathPatch(
                    Path([(x_axis[1], list[1]), (x_axis[1]+0.5, list[2]), (x_axis[2] ,list[2])],
                         [Path.MOVETO, Path.CURVE3, Path.CURVE3]),
                    fc="none", transform=ax1.transData, color=cmap(i))
            ax1.add_patch(path_patch_2)
        if len(list) == 4:
            path_patch_3 = mpatches.PathPatch(
                    Path([(x_axis[2], list[2]), (x_axis[2]+0.5, list[3]), (x_axis[3] ,list[3])],
                         [Path.MOVETO, Path.CURVE3, Path.CURVE3]),
                    fc="none", transform=ax1.transData, color=cmap(i))
            ax1.add_patch(path_patch_3)
        if len(list) == 4:
            ax1.scatter(x_axis,list,color=cmap(i), marker='o',zorder=2,edgecolors= "black",linewidth=0.5)
        elif len(list) == 3:
            ax1.scatter(x_axis_1,list,color=cmap(i), marker='o',zorder=2,edgecolors= "black",linewidth=0.5)
        elif len(list) == 2:
            ax1.scatter(x_axis_2,list,color=cmap(i), marker='o',zorder=2,edgecolors= "black",linewidth=0.5)
        elif len(list) == 1:
            ax1.scatter(x_axis_3,list,color=cmap(i), marker='o',zorder=2,edgecolors= "black",linewidth=0.5)


    plt.xticks(range(0,1), x_axis_names)
    if type_min is not None:
        plt.xticks(range(0,2), x_axis_names)
    if lot is not None:
        plt.xticks(range(0,3), x_axis_names)
    if lot_sp is not None:
        plt.xticks(range(0,4), x_axis_names)

    if len(energy_min_dft_sp) != 0:
        textstr = r'{0} = {1} $\pm$ {2} (kcal/mol)'.format(x_axis_names_mae[0], round(mae_rdkit, 2),round(sd_rdkit, 2))+'\n'
        textstr += r'{0} = {1} $\pm$ {2} (kcal/mol)'.format(x_axis_names_mae[1],round(mae_min,2),round(sd_min,2)) +'\n'
        textstr += r'{0} = {1} $\pm$ {2} (kcal/mol)'.format(x_axis_names_mae[2],round(mae_dft,2),round(sd_dft,2))
        plt.figtext(0.5, -0.05, textstr ,  ha="center", fontsize=12,bbox=dict(facecolor='grey', alpha=0.25))
    elif len(energy_min_dft) != 0:
        textstr = r'{0} = {1} $\pm$ {2} (kcal/mol)'.format(x_axis_names_mae[0], round(mae_rdkit, 2),round(sd_rdkit, 2))+'\n'
        textstr += r'{0} = {1} $\pm$ {2} (kcal/mol)'.format(x_axis_names_mae[1],round(mae_min,2),round(sd_min,2))
        plt.figtext(0.5, -0.03, textstr ,  ha="center", fontsize=12,bbox=dict(facecolor='grey', alpha=0.25))
    #ax1.legend(lines,labels,loc='upper center', prop={'size':4}, bbox_to_anchor=(0.5, -0.13), fancybox=True, shadow=True, ncol=5)
    ax1.set_xlabel('Type of Calculation',fontsize=10)
    ax1.set_ylabel('Relative Energy (kcal/mol)',fontsize=10)
    #plt.setp(ax1.get_xticklabels(), rotation=60, ha="right", visible=True)
    plt.grid(linestyle='--', linewidth=0.75)
    plt.setp(ax1.get_xticklabels(), rotation=0, visible=True)
    title_string=('Energies of Conformers for different Methods : {0}'.format(name_mol))
    ax1.set_title(title_string, fontsize=12)
    fig.tight_layout()
    fig.subplots_adjust(top=0.92,bottom=0.2)
    #creating folder for all molecules to write geom parameter
    folder = w_dir_initial + '/QSTAT/graph'
    try:
        os.makedirs(folder)
        os.chdir(folder)
    except OSError:
        if os.path.isdir(folder):
            os.chdir(folder)
        else:
            raise

    name_to_write = type_csearch
    if type_min is not None:
        name_to_write += '-'+type_min
    if lot is not None:
        name_to_write += '-'+lot+'-'+bs
    if lot_sp is not None:
        name_to_write += '-'+lot_sp+'-'+bs_sp
    plt.savefig(name_mol+'-'+name_to_write+'.png',bbox_inches='tight', format='png', dpi=400)
    plt.close()
    os.chdir(w_dir_initial)

def graph(sdf_rdkit,sdf_xtb,sdf_ani,qm_files,sp_files,args,log,lot,bs,lot_sp,bs_sp,name_mol,w_dir_initial,w_dir_sp,w_dir,type):

    try:
        import cclib
    except (ModuleNotFoundError,AttributeError):
        log.write('The cclib module is not installed correctly - the graph funtcion is not available')

    inmols_rdkit = Chem.SDMolSupplier(sdf_rdkit, removeHs=False)
    #get the energy from sdf
    energy_rdkit = get_energy(inmols_rdkit)
    energy_rdkit_sc = scaling_with_lowest(energy_rdkit)

    if os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf'):
        #get energy list for all conformers from sdfs of rdkit and minimize
        inmols_xtb =  Chem.SDMolSupplier(sdf_xtb, removeHs=False)
        energy_xtb = get_energy(inmols_xtb)
        energy_xtb = rename_name(energy_xtb,'xtb')
        energy_xtb_sc = scaling_with_lowest(energy_xtb)
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf'):
        #get energy list for all conformers from sdfs of rdkit and minimize
        inmols_ani = Chem.SDMolSupplier(sdf_ani, removeHs=False)
        energy_ani = get_energy(inmols_ani)
        energy_ani = rename_name(energy_ani,'ani')
        energy_ani_sc = scaling_with_lowest(energy_ani)

    energy_rdkit_dft,energy_xtb_dft,energy_ani_dft = [],[],[]
    sp_rdkit_dft,sp_xtb_dft,sp_ani_dft = [],[],[]
    #get energy from log FILES
    if qm_files is not None:
        for file in qm_files:
            data = cclib.io.ccread(file)
            if len(file.split('_ani.log')) == 2 or len(file.split('_xtb.log')) == 2:
                if len(file.split('_ani.log')) == 2:
                    name = file.split('_ani.log')[0]
                    energy_ani_dft.append([name,data.scfenergies[0]*ev_2_kcal_mol])
                if len(file.split('_xtb.log')) == 2:
                    name = file.split('_xtb.log')[0]
                    energy_xtb_dft.append([name,data.scfenergies[0]*ev_2_kcal_mol])
            else:
                name = file.split('.log')[0]
                energy_rdkit_dft.append([name,data.scfenergies[0]*ev_2_kcal_mol])

    if sp_files is not None:
        os.chdir(w_dir_sp)
        if type=='g16':
            for file in sp_files:
                data_sp = cclib.io.ccread(file)
                if len(file.split('_ani_'+args.suffix+'.log')) == 2 or len(file.split('_xtb_'+args.suffix+'.log')) == 2:
                    if len(file.split('_ani_'+args.suffix+'.log')) == 2:
                        name = file.split('_ani_'+args.suffix+'.log')[0]
                        sp_ani_dft.append([name,data_sp.scfenergies[0]*ev_2_kcal_mol])
                    if len(file.split('_xtb_'+args.suffix+'.log')) == 2:
                        name = file.split('_xtb_'+args.suffix+'.log')[0]
                        sp_xtb_dft.append([name,data_sp.scfenergies[0]*ev_2_kcal_mol])
                else:
                    name = file.split('.log')[0]
                    sp_rdkit_dft.append([name,data_sp.scfenergies[0]*ev_2_kcal_mol])
        if type=='orca':
            for file in sp_files:
                sp_lines = open(file,"r").readlines()
                for i,line in enumerate(sp_lines):
                    if sp_lines[i].find('FINAL SINGLE POINT ENERGY') > -1:
                        if len(file.split('_ani_'+args.suffix+'.out')) == 2 or len(file.split('_xtb_'+args.suffix+'.out')) == 2:
                            if len(file.split('_ani_'+args.suffix+'.out')) == 2:
                                name = file.split('_ani_'+args.suffix+'.out')[0]
                                sp_ani_dft.append([name,float(sp_lines[i].split()[-1])*hartree_to_kcal])
                            if len(file.split('_xtb_'+args.suffix+'.out')) == 2:
                                name = file.split('_xtb_'+args.suffix+'.out')[0]
                                sp_xtb_dft.append([name,float(sp_lines[i].split()[-1])*hartree_to_kcal])
                        else:
                            name = file.split('.out')[0]
                            sp_rdkit_dft.append([name,float(sp_lines[i].split()[-1])*hartree_to_kcal])
    os.chdir(w_dir)


    energy_ani_dft_sc, energy_xtb_dft_sc, energy_rdkit_dft_sc = [],[],[]
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and qm_files is not None:
        energy_ani_dft_sc = scaling_with_lowest(energy_ani_dft)
    if os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and qm_files is not None:
        energy_xtb_dft_sc = scaling_with_lowest(energy_xtb_dft)
    if not os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and qm_files is not None:
        energy_rdkit_dft_sc = scaling_with_lowest(energy_rdkit_dft)


    sp_ani_dft_sc, sp_xtb_dft_sc, sp_rdkit_dft_sc = [],[],[]
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and sp_files is not None:
        sp_ani_dft_sc = scaling_with_lowest(sp_ani_dft)
    if os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and sp_files is not None:
        sp_xtb_dft_sc = scaling_with_lowest(sp_xtb_dft)
    if not os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and sp_files is not None:
        sp_rdkit_dft_sc = scaling_with_lowest(sp_rdkit_dft)


    if  os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/rdkit/'+name_mol+'_rdkit.sdf'):
        plot_graph(energy_rdkit_sc,energy_xtb_sc,energy_xtb_dft_sc,lot,bs,sp_xtb_dft_sc,lot_sp,bs_sp,name_mol,args,log,'RDKit','xTB',w_dir_initial)
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/rdkit/'+name_mol+'_rdkit.sdf'):
        plot_graph(energy_rdkit_sc,energy_ani_sc,energy_ani_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'RDKit','ANI',w_dir_initial)

    if  os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/summ/'+name_mol+'_summ.sdf'):
        plot_graph(energy_rdkit_sc,energy_xtb_sc,energy_xtb_dft_sc,lot,bs,sp_xtb_dft_sc,lot_sp,bs_sp,name_mol,args,log,'SUMM','xTB',w_dir_initial)
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/summ/'+name_mol+'_summ.sdf'):
        plot_graph(energy_rdkit_sc,energy_ani_sc,energy_ani_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'SUMM','ANI',w_dir_initial)

    if  os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/fullmonte/'+name_mol+'_fullmonte.sdf'):
        plot_graph(energy_rdkit_sc,energy_xtb_sc,energy_xtb_dft_sc,lot,bs,sp_xtb_dft_sc,lot_sp,bs_sp,name_mol,args,log,'Fullmonte','xTB',w_dir_initial)
    if os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') and os.path.exists(w_dir_initial+'/CSEARCH/fullmonte/'+name_mol+'_fullmonte.sdf'):
        plot_graph(energy_rdkit_sc,energy_ani_sc,energy_ani_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'Fullmonte','ANI',w_dir_initial)

    if os.path.exists(w_dir_initial+'/CSEARCH/summ/'+name_mol+'_summ.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf'):
        plot_graph(energy_rdkit_sc,None,energy_rdkit_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'SUMM',None,w_dir_initial)
    if os.path.exists(w_dir_initial+'/CSEARCH/rdkit/'+name_mol+'_rdkit.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') :
        plot_graph(energy_rdkit_sc,None,energy_rdkit_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'RDKit',None,w_dir_initial)
    if os.path.exists(w_dir_initial+'/CSEARCH/fullmonte/'+name_mol+'_fullmonte.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/xtb/'+name_mol+'_xtb.sdf') and not os.path.exists(w_dir_initial+'/CSEARCH/ani/'+name_mol+'_ani.sdf') :
        plot_graph(energy_rdkit_sc,None,energy_rdkit_dft_sc,lot,bs,sp_ani_dft_sc,lot_sp,bs_sp,name_mol,args,log,'Fullmonte',None,w_dir_initial)



from setuptools import setup, find_packages
import io

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name='aqme',
  packages=find_packages(exclude=['tests']),
  package_data={'aqme': ['templates/*']},
  version='0.1',
  license='MIT',
  description='Automated Quantum Mechanical Environments',
  long_description='Automated Quantum Mechanical Environments',
  long_description_content_type='text/markdown',
  author='Shree Sowndarya S. V., Juan V. Alegre Requena',
  author_email='svss@colostate.edu, juanvi89@hotmail.com',
  keywords=['workflows','computational chemistry','conformational sampling','cheminformatics','quantum mechanics','DFT','automated'],
  url = 'https://github.com/jvalegre/aqme',
  download_url = 'https://github.com/jvalegre/aqme/archive/refs/tags/0.1.tar.gz',
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',      #Specify which python versions you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],
  install_requires=['PyYAML>=5.3',
    'pandas>=1.0.3',
    'progress>=1.5',
    'ase>=3.19.1',
    'numpy>=1.18.4',
    'cclib>=1.6.3',
    'matplotlib>=3.2.2',
    'seaborn>=0.10.1',
    'torch>=1.5.0+cpu',
    'torchvision>=0.6.0+cpu',
    'torchani>=2.2'],
  python_requires='>=3.0',
  include_package_data=True,
)
